import contextlib
import io
import json
import os
import sys
import textwrap

from peru.async_helpers import raises_gathered
import peru.cache
import peru.compat
import peru.error
import peru.main
from peru.parser import DEFAULT_PERU_FILE_NAME
import peru.rule
import peru.scope

import shared
from shared import run_peru_command, assert_contents

PERU_MODULE_ROOT = os.path.abspath(
    os.path.join(os.path.dirname(peru.__file__)))


class SyncTest(shared.PeruTest):
    def setUp(self):
        self.test_dir = shared.create_dir()
        self.peru_dir = os.path.join(self.test_dir, '.peru')

    def tearDown(self):
        shared.assert_clean_tmp(self.peru_dir)

    def write_yaml(self, unformatted_yaml, *format_args, dir=None):
        yaml = textwrap.dedent(unformatted_yaml.format(*format_args))
        if dir is None:
            dir = self.test_dir
        with open(os.path.join(dir, DEFAULT_PERU_FILE_NAME), 'w') as f:
            f.write(yaml)

    def do_integration_test(self,
                            args,
                            expected,
                            *,
                            cwd=None,
                            **peru_cmd_kwargs):
        if not cwd:
            cwd = self.test_dir
        output = run_peru_command(args, cwd, **peru_cmd_kwargs)
        assert_contents(
            self.test_dir,
            expected,
            excludes=[DEFAULT_PERU_FILE_NAME, '.peru'])
        return output

    def test_basic_sync(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {}

            imports:
                foo: subdir
            ''', module_dir)
        self.do_integration_test(['sync'], {'subdir/foo': 'bar'})

        # Running it again should be a no-op.
        self.do_integration_test(['sync'], {'subdir/foo': 'bar'})

        # Running it with a dirty working copy should be an error.
        shared.write_files(self.test_dir, {'subdir/foo': 'dirty'})
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            self.do_integration_test(['sync'], {'subdir/foo': 'bar'})

    def test_no_cache_flag(self):
        foo_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {}

            imports:
                foo: subdir
            ''', foo_dir)

        # Sync the foo module once.
        self.do_integration_test(['sync'], {'subdir/foo': 'bar'})

        # Change the contents of foo and sync again. Because foo is cached, we
        # shouldn't see any changes.
        shared.write_files(foo_dir, {'foo': 'woo'})
        self.do_integration_test(['sync'], {'subdir/foo': 'bar'})

        # Now sync with --no-cache. This time we should see the changes.
        self.do_integration_test(['sync', '--no-cache'], {'subdir/foo': 'woo'})

    def test_sync_from_subdir(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            # Use a relative module path, to make sure it gets resolved
            # relative to the project root and not the dir where peru was
            # called.
            cp module relative_foo:
                path: {}

            imports:
                relative_foo: subdir
            ''', os.path.relpath(module_dir, start=self.test_dir))
        subdir = os.path.join(self.test_dir, 'a', 'b')
        peru.compat.makedirs(subdir)
        run_peru_command(['sync'], subdir)
        self.assertTrue(
            os.path.isdir(os.path.join(self.test_dir, '.peru')),
            msg=".peru dir didn't end up in the right place")
        assert_contents(os.path.join(self.test_dir, 'subdir'), {'foo': 'bar'})

    def test_conflicting_imports(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {0}

            # same as foo
            cp module bar:
                path: {0}

            imports:
                foo: subdir
                bar: subdir
            ''', module_dir)
        with self.assertRaises(peru.cache.MergeConflictError):
            self.do_integration_test(['sync'], {'subdir/foo': 'bar'})

    def test_empty_imports(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        empty_yaml = '''\
            cp module foo:
                path: {}
            '''.format(module_dir)
        nonempty_yaml = '''\
            cp module foo:
                path: {}
            imports:
                foo: ./
            '''.format(module_dir)
        self.write_yaml(empty_yaml)
        self.do_integration_test(['sync'], {})
        # Now test switching back and forth between non-empty and empty.
        self.write_yaml(nonempty_yaml)
        self.do_integration_test(['sync'], {'foo': 'bar'})
        # Back to empty.
        self.write_yaml(empty_yaml)
        self.do_integration_test(['sync'], {})

    def test_import_module_defined_in_another_module(self):
        # Project B contains project A
        dir_a = shared.create_dir({'afile': 'stuff'})
        dir_b = shared.create_dir()
        # Create the peru.yaml file for B.
        self.write_yaml(
            '''\
            cp module a:
                path: {}
            ''',
            dir_a,
            dir=dir_b)
        # Now create the peru.yaml file in the actual test project.
        self.write_yaml(
            '''\
            imports:
                b.a: a_via_b/

            cp module b:
                path: {}
            ''', dir_b)
        self.do_integration_test(['sync'], {'a_via_b/afile': 'stuff'})
        # Test the error message from an invalid module.
        self.write_yaml(
            '''\
            imports:
                b.missing_module: some_path

            cp module b:
                path: {}
            ''', dir_b)
        try:
            self.do_integration_test(['sync'], {})
        except peru.error.PrintableError as e:
            assert 'b.missing_module' in e.message
        else:
            assert False, 'should throw invalid module error'

    def test_recursive_imports(self):
        # Project B contains project A
        dir_a = shared.create_dir({'afile': 'aaa'})
        dir_b = shared.create_dir({'exports/bfile': 'bbb'})
        # Create the peru.yaml file for B.
        self.write_yaml(
            '''\
            imports:
                a: exports/where_b_put_a
            cp module a:
                path: {}
            ''',
            dir_a,
            dir=dir_b)
        # Now create the peru.yaml file in the actual test project.
        self.write_yaml(
            '''\
            imports:
                b: where_c_put_b

            cp module b:
                # recursive is false by default
                path: {}
                export: exports  # omit the peru.yaml file from b
            ''', dir_b)
        self.do_integration_test(['sync'], {'where_c_put_b/bfile': 'bbb'})

        # Repeat the same test with explicit 'recursive' settings.
        self.write_yaml(
            '''\
            imports:
                b: where_c_put_b

            cp module b:
                path: {}
                pick: exports/where_b_put_a
                export: exports  # omit the peru.yaml file from b
                recursive: true
            ''', dir_b)
        self.do_integration_test(['sync'],
                                 {'where_c_put_b/where_b_put_a/afile': 'aaa'})

        self.write_yaml(
            '''\
            imports:
                b: where_c_put_b

            cp module b:
                path: {}
                export: exports  # omit the peru.yaml file from b
                recursive: false
            ''', dir_b)
        self.do_integration_test(['sync'], {'where_c_put_b/bfile': 'bbb'})

    def test_recursive_import_error(self):
        '''Errors that happen inside recursively-fetched targets should have
        context information about the targets that caused them. This test is
        especially important for checking that context isn't lost in
        GatheredExceptions.'''
        # Project NOTABLE_NAME has a BAD_MODULE in it.
        dir_notable = shared.create_dir()
        # Create the peru.yaml file for NOTABLE_NAME.
        self.write_yaml(
            '''\
            imports:
                BAD_MODULE: ./
            git module BAD_MODULE:
                bad_field: stuff
                # The error we get here will actually be that `url` is missing.
            ''',
            dir=dir_notable)
        # Now make our test project import it.
        self.write_yaml(
            '''\
            imports:
                NOTABLE_NAME: ./notable

            cp module NOTABLE_NAME:
                recursive: true
                path: {}
            ''', dir_notable)
        with self.assertRaises(peru.error.PrintableError) as cm:
            run_peru_command(['sync'], self.test_dir)
        self.assertIn("NOTABLE_NAME", cm.exception.message)
        self.assertIn("BAD_MODULE", cm.exception.message)

    def test_peru_file_field(self):
        # Project B contains project A
        dir_a = shared.create_dir({'afile': 'stuff'})
        # Create project B with an unusual YAML filename.
        dir_b = shared.create_dir({
            'alternate.yaml':
            textwrap.dedent('''\
            cp module a:
                path: {}
            '''.format(dir_a))
        })
        # Now create the peru.yaml file in the actual test project.
        self.write_yaml(
            '''\
            imports:
                b.a: a_via_b/

            cp module b:
                path: {}
                peru file: alternate.yaml
            ''', dir_b)
        self.do_integration_test(['sync'], {'a_via_b/afile': 'stuff'})

    def test_module_rules(self):
        module_dir = shared.create_dir({'a/b': '', 'c/d': ''})
        yaml = '''\
            cp module foo:
                path: {}

            rule get_a:
                export: a

            rule get_c:
                export: c

            imports:
                foo|get_a: ./
            '''.format(module_dir)
        self.write_yaml(yaml)
        self.do_integration_test(['sync'], {'b': ''})
        # Run it again with a different import to make sure we clean up.
        yaml_different = yaml.replace('foo|get_a', 'foo|get_c')
        self.write_yaml(yaml_different)
        self.do_integration_test(['sync'], {'d': ''})

    def test_rule_with_picked_files(self):
        content = {
            name: ''
            for name in
            ['foo', 'bar', 'special', 'baz/bing', 'baz/boo/a', 'baz/boo/b']
        }
        module_dir = shared.create_dir(content)
        self.write_yaml(
            '''\
            cp module foo:
                path: {}

            rule filter:
                pick:
                  - "**/*oo"
                  - special

            imports:
                foo|filter: ./
            ''', module_dir)
        filtered_content = {
            name: ''
            for name in [
                'foo',
                'special',
                'baz/boo/a',
                'baz/boo/b',
            ]
        }
        self.do_integration_test(['sync'], filtered_content)

    def test_rule_with_picked_files_that_do_not_exist(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {}
                pick: idontexist
            imports:
                foo: ./
            ''', module_dir)
        with raises_gathered(peru.rule.NoMatchingFilesError):
            self.do_integration_test(['sync'], {})

    def test_rule_with_exported_files_that_are_not_picked(self):
        content = {
            name: ''
            for name in ['foo', 'bar', 'baz/bing', 'baz/boo/a', 'baz/boo/b']
        }
        module_dir = shared.create_dir(content)
        self.write_yaml(
            '''\
            cp module foo:
                path: {}
                pick: foo
                export: baz/

            imports:
                foo: ./
            ''', module_dir)
        with raises_gathered(peru.rule.NoMatchingFilesError):
            self.do_integration_test(['sync'], {})

    def test_rule_with_dropped_files(self):
        content = {'foo': 'one', 'bar': 'two'}
        module_dir = shared.create_dir(content)
        self.write_yaml(
            '''\
            cp module foobar:
                path: {}

            rule filter:
                drop: foo

            imports:
                foobar|filter: ./
            ''', module_dir)
        filtered_content = {'bar': 'two'}
        self.do_integration_test(['sync'], filtered_content)

    def test_drop_then_pick_is_an_error(self):
        '''We want drop to run before pick, so that deleting a bunch of stuff
        and then trying to pick it turns into an error. The opposite execution
        order would make this silently succeed. See the discussion at
        https://github.com/buildinspace/peru/issues/150#issuecomment-212580912.
        '''
        content = {'foo': 'stuff'}
        module_dir = shared.create_dir(content)
        self.write_yaml(
            '''\
            cp module foobar:
                path: {}
                drop: foo
                pick: foo

            imports:
                foobar: ./
            ''', module_dir)
        with raises_gathered(peru.rule.NoMatchingFilesError):
            run_peru_command(['sync'], self.test_dir)

    def test_rule_with_executable(self):
        contents = {'a.txt': '', 'b.txt': '', 'c.foo': ''}
        module_dir = shared.create_dir(contents)
        self.write_yaml(
            '''\
            cp module foo:
                path: {}
                executable: "*.txt"
            imports:
                foo: ./
            ''', module_dir)
        self.do_integration_test(['sync'], contents)
        for f in ('a.txt', 'b.txt'):
            shared.assert_executable(os.path.join(self.test_dir, f))

    def test_rule_with_move(self):
        module_dir = shared.create_dir({'a': 'foo', 'b/c': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {}
                move:
                    a: newa
                    b: newb
            imports:
                foo: ./
            ''', module_dir)
        self.do_integration_test(['sync'], {'newa': 'foo', 'newb/c': 'bar'})

    def test_rule_with_move_error(self):
        module_dir = shared.create_dir()
        self.write_yaml(
            '''\
            cp module foo:
                path: {}
                move:
                    doesntexist: also_nonexistent
            imports:
                foo: ./
            ''', module_dir)
        with raises_gathered(peru.rule.NoMatchingFilesError) as cm:
            self.do_integration_test(['sync'], {
                'newa': 'foo',
                'newb/c': 'bar'
            })
        assert 'doesntexist' in cm.exception.message

    def test_rule_with_copied_files(self):
        content = {'foo': 'foo', 'bar/baz': 'baz'}
        module_dir = shared.create_dir(content)
        self.write_yaml(
            '''\
            cp module foo:
                path: {}
                copy:
                    foo: foo-copy
                    bar:
                      - bar-copy-1
                      - bar-copy-2

            imports:
                foo: ./
            ''', module_dir)
        copied_content = {
            'foo': 'foo',
            'bar/baz': 'baz',
            'foo-copy': 'foo',
            'bar-copy-1/baz': 'baz',
            'bar-copy-2/baz': 'baz'
        }
        self.do_integration_test(['sync'], copied_content)

    def test_alternate_cache(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {}

            imports:
                foo: subdir
            ''', module_dir)
        cache_dir = shared.create_dir()
        env = {'PERU_CACHE_DIR': cache_dir}
        self.do_integration_test(['sync'], {'subdir/foo': 'bar'}, env=env)
        self.assertTrue(os.path.exists(os.path.join(cache_dir, 'plugins')))
        self.assertTrue(os.path.exists(os.path.join(cache_dir, 'trees')))
        self.assertTrue(os.path.exists(os.path.join(cache_dir, 'keyval')))
        self.assertFalse(os.path.exists(os.path.join(self.peru_dir, 'cache')))

    def test_override(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {}

            imports:
                foo: ./
            ''', module_dir)
        override_dir = shared.create_dir({'foo': 'override'})
        # Set the override.
        run_peru_command(['override', 'add', 'foo', override_dir],
                         self.test_dir)
        # Confirm that the override is configured.
        output = run_peru_command(['override'], self.test_dir)
        self.assertEqual(output, 'foo: {}\n'.format(override_dir))
        # Make sure 'override list' gives the same output as 'override'.
        output = run_peru_command(['override', 'list'], self.test_dir)
        self.assertEqual(output, 'foo: {}\n'.format(override_dir))
        # Same as above, but as JSON (with --json flag).
        output = run_peru_command(['override', '--json'], self.test_dir)
        override_dict = json.loads(output)
        self.assertEqual(override_dict, {'foo': override_dir})
        # Run the sync with --no-overrides and confirm nothing changes. Also
        # check that there's no overrides-related output.
        output = self.do_integration_test(['sync', '--no-overrides'],
                                          {'foo': 'bar'})
        self.assertNotIn('overrides', output)
        # Now run the sync normally and confirm that the override worked. Also
        # confirm that we mentioned the override in output, and that the unused
        # overrides warning is not printed.
        output = self.do_integration_test(['sync'], {'foo': 'override'})
        self.assertIn('overrides', output)
        self.assertNotIn('WARNING unused overrides', output)
        # Delete the override.
        run_peru_command(['override', 'delete', 'foo'], self.test_dir)
        # Confirm that the override was deleted.
        output = run_peru_command(['override'], self.test_dir)
        self.assertEqual(output, '')
        # Rerun the sync and confirm the original content is back.
        self.do_integration_test(['sync'], {'foo': 'bar'})
        # Add a bogus override and confirm the unused overrides warning is
        # printed.
        run_peru_command(['override', 'add', 'bogus', override_dir],
                         self.test_dir)
        output = self.do_integration_test(['sync'], {'foo': 'bar'})
        self.assertIn('WARNING unused overrides', output)

    def test_override_after_regular_sync(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {}

            imports:
                foo: ./
            ''', module_dir)
        # First, do a regular sync.
        self.do_integration_test(['sync'], {'foo': 'bar'})
        # Now, add an override, and confirm that the new sync works.
        override_dir = shared.create_dir({'foo': 'override'})
        run_peru_command(['override', 'add', 'foo', override_dir],
                         self.test_dir)
        self.do_integration_test(['sync'], {'foo': 'override'})

    def test_override_recursive(self):
        # Module A just includes the file 'foo'.
        module_a_dir = shared.create_dir({'foo': 'bar'})
        # Module B imports module A.
        module_b_dir = shared.create_dir()
        self.write_yaml(
            '''\
            cp module A:
                path: {}

            imports:
                A: A/
            ''',
            module_a_dir,
            dir=module_b_dir)
        # Module C (in self.test_dir) imports module B, and also directly
        # imports module A. When we set an override for module A below, we'll
        # want to check that *both* of these imports get overridden.
        self.write_yaml(
            '''\
            cp module B:
                path: {}
                recursive: true
                # Note that module business happens before rule business, so
                # 'drop: peru.yaml' will not affect the recursion, just the
                # final output.
                drop: peru.yaml

            imports:
                B.A: A/
                B: B/
            ''', module_b_dir)
        # First, do a regular sync.
        self.do_integration_test(['sync'], {
            'A/foo': 'bar',
            'B/A/foo': 'bar',
        })
        # Now set an override for B.A.
        override_dir = shared.create_dir({'foo': 'override'})
        run_peru_command(['override', 'add', 'B.A', override_dir],
                         self.test_dir)
        # Now do another sync. *Both* the directly imported copy of A *and* the
        # copy synced inside of B should be overridden.
        self.do_integration_test(['sync'], {
            'A/foo': 'override',
            'B/A/foo': 'override',
        })

    def test_relative_override_from_subdir(self):
        self.write_yaml('''\
            empty module foo:

            imports:
                foo: ./
            ''')
        # Create some subdirs inside the project.
        subdir = os.path.join(self.test_dir, 'a', 'b')
        peru.compat.makedirs(subdir)
        # Create an override dir outside the project.
        override_dir = shared.create_dir({'foo': 'override'})
        # Set the override from inside subdir, using the relative path that's
        # valid from that location. Peru is going to store this path in
        # .peru/overrides/ at the root, so this tests that we resolve the
        # stored path properly.
        relative_path = os.path.relpath(override_dir, start=subdir)
        run_peru_command(['override', 'add', 'foo', relative_path], subdir)
        # Confirm that the right path is stored on disk.
        expected_stored_path = os.path.relpath(
            override_dir, start=self.test_dir)
        with open(os.path.join(self.peru_dir, 'overrides', 'foo')) as f:
            actual_stored_path = f.read()
        self.assertEqual(expected_stored_path, actual_stored_path)
        # Confirm that `peru override` prints output that respects the cwd.
        output = run_peru_command(['override'], subdir)
        self.assertEqual('foo: {}\n'.format(relative_path), output)
        # Confirm that syncing works.
        self.do_integration_test(['sync'], {'foo': 'override'}, cwd=subdir)

    def test_override_excludes_dotperu(self):
        self.write_yaml('''\
            empty module foo:

            imports:
                foo: ./
            ''')
        override_dir = shared.create_dir({
            'foo': 'override',
            '.peru/bar': 'baz'
        })
        run_peru_command(['override', 'add', 'foo', override_dir],
                         self.test_dir)
        self.do_integration_test(['sync'], {'foo': 'override'})

    def test_rules_in_override(self):
        module_dir = shared.create_dir({'a/b': 'c'})
        yaml = '''
            imports:
                foo|get_a: ./

            cp module foo:
                path: {}

            rule get_a:
                export: a
            '''
        self.write_yaml(yaml, module_dir)
        override_dir = shared.create_dir({'a/b': 'override'})
        run_peru_command(['override', 'add', 'foo', override_dir],
                         self.test_dir)
        self.do_integration_test(['sync'], {'b': 'override'})

    def test_missing_name_errors(self):
        self.write_yaml('''
            imports:
                thingabc: path
            ''')
        with self.assertRaises(peru.error.PrintableError) as cm:
            self.do_integration_test(['sync'], {})
        assert "thingabc" in cm.exception.message
        self.write_yaml('''
            imports:
                thingabc|rulexyz: path

            git module thingabc:
                url: http://example.com
            ''')
        with self.assertRaises(peru.error.PrintableError) as cm:
            self.do_integration_test(['sync'], {})
        assert "rulexyz" in cm.exception.message, cm.exception.message

    def test_copy(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {}
            ''', module_dir)
        # Do a simple copy and check the results.
        self.do_integration_test(['copy', 'foo', '.'], {'foo': 'bar'})
        # Running the same copy again should fail, because of conflicts.
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            self.do_integration_test(['copy', 'foo', '.'], {'foo': 'bar'})
        # Passing the --force flag should pave over conflicts.
        self.do_integration_test(['copy', '--force', 'foo', '.'],
                                 {'foo': 'bar'})

    def test_copy_nested(self):
        # Project B contains project A
        dir_a = shared.create_dir({'afile': 'stuff'})
        dir_b = shared.create_dir()
        # Create the peru.yaml file for B.
        self.write_yaml(
            '''\
            cp module a:
                path: {}
            ''',
            dir_a,
            dir=dir_b)
        # Now create the peru.yaml file in the actual test project.
        self.write_yaml(
            '''\
            cp module b:
                path: {}
            ''', dir_b)
        self.do_integration_test(['copy', 'b.a', '.'], {'afile': 'stuff'})

    def test_clean(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            imports:
                foo: ./
            cp module foo:
                path: {}
            ''', module_dir)
        self.do_integration_test(['clean'], {})
        self.do_integration_test(['sync'], {'foo': 'bar'})
        shared.write_files(self.test_dir, {'foo': 'DIRTY'})
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            self.do_integration_test(['clean'], {})
        self.do_integration_test(['clean', '--force'], {})

    def test_help(self):
        flag_output = run_peru_command(['--help'], self.test_dir)
        self.assertEqual(peru.main.__doc__, flag_output)

        command_output = run_peru_command(['help'], self.test_dir)
        self.assertEqual(peru.main.__doc__, command_output)

        clean_help = peru.main.COMMAND_DOCS['clean']

        pre_flag_output = run_peru_command(['-h', 'clean'], self.test_dir)
        self.assertEqual(clean_help, pre_flag_output)

        post_flag_output = run_peru_command(['clean', '-h'], self.test_dir)
        self.assertEqual(clean_help, post_flag_output)

        buffer = io.StringIO()
        with redirect_stderr(buffer):
            run_peru_command(['foobarbaz'], self.test_dir, expected_error=1)
        self.assertEqual(peru.main.__doc__, buffer.getvalue())

    def test_version(self):
        version_output = run_peru_command(["--version"], self.test_dir)
        self.assertEqual(peru.main.get_version(), version_output.strip())

    def test_duplicate_keys_warning(self):
        self.write_yaml('''\
            git module foo:
            git module foo:
            ''')
        buffer = io.StringIO()
        with redirect_stderr(buffer):
            run_peru_command(['sync'], self.test_dir)
        assert ('WARNING' in buffer.getvalue())
        assert ('git module foo' in buffer.getvalue())
        # Make sure --quiet suppresses the warning.
        buffer = io.StringIO()
        with redirect_stderr(buffer):
            run_peru_command(['sync', '--quiet'], self.test_dir)
        # Don't literally check that stderr is empty, because that could get
        # tripped up on other Python warnings (like asyncio taking too long).
        assert 'git module foo' not in buffer.getvalue()

    def test_lastimports_timestamp(self):
        module_dir = shared.create_dir({'foo': 'bar'})
        template = '''\
            cp module foo:
                path: {}

            imports:
                foo: {}
            '''
        self.write_yaml(template, module_dir, "subdir1")
        self.do_integration_test(['sync'], {'subdir1/foo': 'bar'})
        lastimports = os.path.join(self.test_dir, '.peru', 'lastimports')

        def get_timestamp():
            return os.stat(lastimports).st_mtime

        original_timestamp = get_timestamp()

        # Running it again should be a no-op. Assert that the lastimports
        # timestamp hasn't changed.
        self.do_integration_test(['sync'], {'subdir1/foo': 'bar'})
        assert get_timestamp() == original_timestamp, \
            "Expected an unchanged timestamp."

        # Modify peru.yaml and sync again. This should change the timestamp.
        self.write_yaml(template, module_dir, "subdir2")
        self.do_integration_test(['sync'], {'subdir2/foo': 'bar'})
        assert get_timestamp() > original_timestamp, \
            "Expected an updated timestamp."

    def test_number_of_git_commands(self):
        '''A no-op sync should be a single git command. Also check that index
        files are deleted after any sync error.'''
        module_dir = shared.create_dir({'foo': 'bar'})
        self.write_yaml(
            '''\
            cp module foo:
                path: {}

            imports:
                foo: subdir
            ''', module_dir)
        index_path = os.path.join(self.test_dir, '.peru/lastimports.index')

        # The first sync should take multiple operations and create a
        # lastimports.index file.
        peru.cache.DEBUG_GIT_COMMAND_COUNT = 0
        self.do_integration_test(['sync'], {'subdir/foo': 'bar'})
        assert peru.cache.DEBUG_GIT_COMMAND_COUNT > 1, \
            'The first sync should take multiple operations.'
        assert os.path.exists(index_path), \
            'The first sync should create an index file.'

        # The second sync should reuse the index file and only take one
        # operation.
        peru.cache.DEBUG_GIT_COMMAND_COUNT = 0
        self.do_integration_test(['sync'], {'subdir/foo': 'bar'})
        assert peru.cache.DEBUG_GIT_COMMAND_COUNT == 1, \
            'The second sync should take only one operation.'
        assert os.path.exists(index_path), \
            'The second sync should preserve the index file.'

        # Now force an error. This should delete the index file.
        with open(os.path.join(self.test_dir, 'subdir/foo'), 'w') as f:
            f.write('dirty')
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            run_peru_command(['sync'], self.test_dir)
        assert not os.path.exists(index_path), \
            'The error should delete the index file.'

        # Fix the error and resync with new module contents. This should
        # recreate the index file with the current tree and then succeed,
        # rather than using an empty index and treating the current files as
        # conflicting.
        with open(os.path.join(self.test_dir, 'subdir/foo'), 'w') as f:
            f.write('bar')
        with open(os.path.join(module_dir, 'foo'), 'w') as f:
            f.write('new bar')
        self.do_integration_test(['sync', '--no-cache'],
                                 {'subdir/foo': 'new bar'})
        assert os.path.exists(index_path), \
            'The index should have been recreated.'

    def test_module_list(self):
        self.write_yaml('''\
            git module foo:
                url: blah
            git module bar:
                url: blah
            ''')

        output = run_peru_command(['module'], self.test_dir)
        self.assertEqual(output, "bar\nfoo\n")

        output = run_peru_command(['module', 'list'], self.test_dir)
        self.assertEqual(output, "bar\nfoo\n")

        output = run_peru_command(['module', 'list', '--json'], self.test_dir)
        self.assertEqual(output, '["bar", "foo"]\n')


@contextlib.contextmanager
def redirect_stderr(f):
    old_stderr = sys.stderr
    sys.stderr = f
    try:
        yield
    finally:
        sys.stderr = old_stderr



import os

from peru import cache
from peru import rule

import shared
from shared import COLON


class RuleTest(shared.PeruTest):
    @shared.make_synchronous
    async def setUp(self):
        self.cache_dir = shared.create_dir()
        self.cache = await cache.Cache(self.cache_dir)
        # Include a leading colon to test that we prepend ./ to pathspecs.
        self.content = {'a': 'foo', 'b/c': 'bar', COLON + 'd': 'baz'}
        self.content_dir = shared.create_dir(self.content)
        self.content_tree = await self.cache.import_tree(self.content_dir)
        self.entries = await self.cache.ls_tree(
            self.content_tree, recursive=True)

    @shared.make_synchronous
    async def test_copy(self):
        # A file copied into a directory should be placed into that directory.
        # A directory or file copied into a file should overwrite that file.
        copies = {'a': ['x', 'b', 'b/c'], 'b': ['a', 'y']}
        tree = await rule.copy_files(self.cache, self.content_tree, copies)
        await shared.assert_tree_contents(
            self.cache, tree, {
                'a/c': 'bar',
                'b/a': 'foo',
                'b/c': 'foo',
                'x': 'foo',
                'y/c': 'bar',
                COLON + 'd': 'baz',
            })

    @shared.make_synchronous
    async def test_move(self):
        # Same semantics as copy above. Also, make sure that move deletes move
        # sources, but does not delete sources that were overwritten by the
        # target of another move.
        moves = {'a': 'b', 'b': 'a'}
        tree = await rule.move_files(self.cache, self.content_tree, moves)
        await shared.assert_tree_contents(self.cache, tree, {
            'a/c': 'bar',
            'b/a': 'foo',
            COLON + 'd': 'baz',
        })

    @shared.make_synchronous
    async def test_drop(self):
        drop_dir = await rule.drop_files(self.cache, self.content_tree, ['b'])
        await shared.assert_tree_contents(self.cache, drop_dir, {
            'a': 'foo',
            COLON + 'd': 'baz'
        })

        drop_file = await rule.drop_files(self.cache, self.content_tree, ['a'])
        await shared.assert_tree_contents(self.cache, drop_file, {
            'b/c': 'bar',
            COLON + 'd': 'baz'
        })

        drop_colon = await rule.drop_files(self.cache, self.content_tree,
                                           [COLON + 'd'])
        await shared.assert_tree_contents(self.cache, drop_colon, {
            'a': 'foo',
            'b/c': 'bar'
        })

        globs = await rule.drop_files(self.cache, self.content_tree,
                                      ['**/c', '**/a'])
        await shared.assert_tree_contents(self.cache, globs,
                                          {COLON + 'd': 'baz'})

    @shared.make_synchronous
    async def test_pick(self):
        pick_dir = await rule.pick_files(self.cache, self.content_tree, ['b'])
        await shared.assert_tree_contents(self.cache, pick_dir, {'b/c': 'bar'})

        pick_file = await rule.pick_files(self.cache, self.content_tree, ['a'])
        await shared.assert_tree_contents(self.cache, pick_file, {'a': 'foo'})

        pick_colon = await rule.pick_files(self.cache, self.content_tree,
                                           [COLON + 'd'])
        await shared.assert_tree_contents(self.cache, pick_colon,
                                          {COLON + 'd': 'baz'})

        globs = await rule.pick_files(self.cache, self.content_tree,
                                      ['**/c', '**/a'])
        await shared.assert_tree_contents(self.cache, globs, {
            'a': 'foo',
            'b/c': 'bar'
        })

    @shared.make_synchronous
    async def test_executable(self):
        exe = await rule.make_files_executable(self.cache, self.content_tree,
                                               ['b/*'])
        new_content_dir = shared.create_dir()
        await self.cache.export_tree(exe, new_content_dir)
        shared.assert_contents(new_content_dir, self.content)
        shared.assert_not_executable(os.path.join(new_content_dir, 'a'))
        shared.assert_executable(os.path.join(new_content_dir, 'b/c'))

    @shared.make_synchronous
    async def test_export(self):
        b = await rule.get_export_tree(self.cache, self.content_tree, 'b')
        await shared.assert_tree_contents(self.cache, b, {'c': 'bar'})



import os

import peru.runtime as runtime

import shared


class RuntimeTest(shared.PeruTest):
    def test_find_peru_file(self):
        test_dir = shared.create_dir({
            'a/find_me': 'junk',
            'a/b/c/junk': 'junk',
        })
        result = runtime.find_project_file(
            os.path.join(test_dir, 'a', 'b', 'c'), 'find_me')
        expected = os.path.join(test_dir, 'a', 'find_me')
        self.assertEqual(expected, result)



import asyncio
from collections import defaultdict
import contextlib
import hashlib
import io
import os
from pathlib import Path
import shutil
import subprocess
import sys
import textwrap
import unittest

from peru.async_helpers import run_task
import peru.plugin as plugin
import shared
from shared import SvnRepo, GitRepo, HgRepo, assert_contents


HG_MINIMUM_PYTHON_VERSION = (3, 6)


class TestDisplayHandle(io.StringIO):
    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


def test_plugin_fetch(context, type, fields, dest):
    handle = TestDisplayHandle()
    run_task(plugin.plugin_fetch(context, type, fields, dest, handle))
    return handle.getvalue()


def test_plugin_get_reup_fields(context, type, fields):
    handle = TestDisplayHandle()
    return run_task(
        plugin.plugin_get_reup_fields(context, type, fields, handle))


class PluginsTest(shared.PeruTest):
    def setUp(self):
        self.content = {"some": "stuff", "foo/bar": "baz"}
        self.content_dir = shared.create_dir(self.content)
        self.cache_root = shared.create_dir()
        self.plugin_context = plugin.PluginContext(
            cwd='.',
            plugin_cache_root=self.cache_root,
            parallelism_semaphore=asyncio.BoundedSemaphore(
                plugin.DEFAULT_PARALLEL_FETCH_LIMIT),
            plugin_cache_locks=defaultdict(asyncio.Lock),
            tmp_root=shared.create_dir())
        plugin.debug_assert_clean_parallel_count()

    def tearDown(self):
        plugin.debug_assert_clean_parallel_count()

    def do_plugin_test(self,
                       type,
                       plugin_fields,
                       expected_content,
                       *,
                       fetch_dir=None):
        fetch_dir = fetch_dir or shared.create_dir()
        output = test_plugin_fetch(self.plugin_context, type, plugin_fields,
                                   fetch_dir)
        assert_contents(fetch_dir, expected_content)
        return output

    def test_git_plugin(self):
        GitRepo(self.content_dir)
        self.do_plugin_test("git", {"url": self.content_dir}, self.content)

    def test_git_default_branch(self):
        GitRepo(self.content_dir, init_default_branch='main')
        self.do_plugin_test("git", {"url": self.content_dir}, self.content)

    def test_empty_git_rev(self):
        empty_dir = shared.create_dir()
        GitRepo(empty_dir)
        self.do_plugin_test('git', {'url': empty_dir}, {})

    @unittest.skipIf(
        sys.version_info < HG_MINIMUM_PYTHON_VERSION,
        "Python too old for hg",
    )
    def test_hg_plugin(self):
        HgRepo(self.content_dir)
        self.do_plugin_test("hg", {"url": self.content_dir}, self.content)

    def test_svn_plugin(self):
        repo = SvnRepo(self.content_dir)
        self.do_plugin_test('svn', {'url': repo.url}, self.content)

    def test_svn_plugin_reup(self):
        repo = SvnRepo(self.content_dir)
        plugin_fields = {'url': repo.url}
        output = test_plugin_get_reup_fields(self.plugin_context, 'svn',
                                             plugin_fields)
        self.assertDictEqual({'rev': '1'}, output)

    def test_git_plugin_with_submodule(self):
        content_repo = GitRepo(self.content_dir)
        # Git has a small bug: The .gitmodules file is always created with "\n"
        # line endings, even on Windows. With core.autocrlf turned on, that
        # causes a warning when the file is added/committed, because those line
        # endings would get replaced with "\r\n" when the file was checked out.
        # We can just turn autocrlf off for this test to silence the warning.
        content_repo.run('git', 'config', 'core.autocrlf', 'false')
        submodule_dir = shared.create_dir({'another': 'file'})
        submodule_repo = GitRepo(submodule_dir)
        content_repo.run('git', 'submodule', 'add', '-q', submodule_dir,
                         'subdir/', env={"GIT_ALLOW_PROTOCOL": "file"})
        content_repo.run('git', 'commit', '-m', 'submodule commit')
        expected_content = self.content.copy()
        expected_content['subdir/another'] = 'file'
        with open(os.path.join(self.content_dir, '.gitmodules')) as f:
            expected_content['.gitmodules'] = f.read()
        self.do_plugin_test('git', {'url': self.content_dir}, expected_content)

        # Now move the submodule forward. Make sure it gets fetched again.
        shared.write_files(submodule_dir, {'more': 'stuff'})
        submodule_repo.run('git', 'add', '-A')
        submodule_repo.run('git', 'commit', '-m', 'more stuff')
        subprocess.check_output(['git', 'pull', '-q'],
                                cwd=os.path.join(self.content_dir, 'subdir'))
        content_repo.run('git', 'commit', '-am', 'submodule update')
        expected_content['subdir/more'] = 'stuff'
        self.do_plugin_test('git', {'url': self.content_dir}, expected_content)

        # Normally when you run `git submodule add ...`, git puts two things in
        # your repo: an entry in .gitmodules, and a commit object at the
        # appropriate path inside your repo. However, it's possible for those
        # two to get out of sync, especially if you use mv/rm on a directory
        # followed by `git add`, instead of the smarter `git mv`/`git rm`. We
        # need to create this condition and check that we then ignore the
        # submodule.
        shutil.rmtree(os.path.join(self.content_dir, 'subdir'))
        content_repo.run('git', 'commit', '-am', 'inconsistent delete')
        del expected_content['subdir/another']
        del expected_content['subdir/more']
        self.do_plugin_test('git', {'url': self.content_dir}, expected_content)

        # Finally, test explicitly disabling submodule fetching. Start by
        # reverting the 'inconsistent delete' commit from above.
        content_repo.run('git', 'revert', '--no-edit', 'HEAD')
        fields = {'url': self.content_dir, 'submodules': 'false'}
        self.do_plugin_test('git', fields, expected_content)

    def test_git_plugin_with_relative_submodule(self):
        content_repo = GitRepo(self.content_dir)
        # Same autocrlf workaround as above.
        content_repo.run('git', 'config', 'core.autocrlf', 'false')

        # Similar to above, but this time we use a relative path.
        submodule_dir = shared.create_dir({'another': 'file'})
        GitRepo(submodule_dir)
        submodule_basename = os.path.basename(submodule_dir)
        relative_path = "../" + submodule_basename
        content_repo.run('git', 'submodule', 'add', '-q', relative_path,
                         'subdir/', env={"GIT_ALLOW_PROTOCOL": "file"})
        content_repo.run('git', 'commit', '-m', 'submodule commit')
        expected_content = self.content.copy()
        expected_content['subdir/another'] = 'file'
        with open(os.path.join(self.content_dir, '.gitmodules')) as f:
            expected_content['.gitmodules'] = f.read()
        self.do_plugin_test('git', {'url': self.content_dir}, expected_content)

    def test_git_plugin_multiple_fetches(self):
        content_repo = GitRepo(self.content_dir)
        head = content_repo.run('git', 'rev-parse', 'HEAD')
        plugin_fields = {"url": self.content_dir, "rev": head}
        output = self.do_plugin_test("git", plugin_fields, self.content)
        self.assertEqual(output.count("git clone"), 1)
        self.assertEqual(output.count("git fetch"), 0)
        # Add a new file to the directory and commit it.
        shared.write_files(self.content_dir, {'another': 'file'})
        content_repo.run('git', 'add', '-A')
        content_repo.run('git', 'commit', '-m', 'committing another file')
        # Refetch the original rev. Git should not do a git-fetch.
        output = self.do_plugin_test("git", plugin_fields, self.content)
        self.assertEqual(output.count("git clone"), 0)
        self.assertEqual(output.count("git fetch"), 0)
        # Not delete the rev field. Git should default to master and fetch.
        del plugin_fields["rev"]
        self.content["another"] = "file"
        output = self.do_plugin_test("git", plugin_fields, self.content)
        self.assertEqual(output.count("git clone"), 0)
        self.assertEqual(output.count("git fetch"), 1)

    @unittest.skipIf(
        sys.version_info < HG_MINIMUM_PYTHON_VERSION,
        "Python too old for hg",
    )
    def test_hg_plugin_multiple_fetches(self):
        content_repo = HgRepo(self.content_dir)
        head = content_repo.run('hg', 'identify', '--debug', '-r',
                                '.').split()[0]
        plugin_fields = {'url': self.content_dir, 'rev': head}
        output = self.do_plugin_test('hg', plugin_fields, self.content)
        self.assertEqual(output.count('hg clone'), 1)
        self.assertEqual(output.count('hg pull'), 0)
        # Add a new file to the directory and commit it.
        shared.write_files(self.content_dir, {'another': 'file'})
        content_repo.run('hg', 'commit', '-A', '-m', 'committing another file')
        # Refetch the original rev. Hg should not do a pull.
        output = self.do_plugin_test('hg', plugin_fields, self.content)
        self.assertEqual(output.count('hg clone'), 0)
        self.assertEqual(output.count('hg pull'), 0)
        # Not delete the rev field. Git should default to master and fetch.
        del plugin_fields['rev']
        self.content['another'] = 'file'
        output = self.do_plugin_test('hg', plugin_fields, self.content)
        self.assertEqual(output.count('hg clone'), 0)
        self.assertEqual(output.count('hg pull'), 1)

    def test_git_plugin_reup(self):
        repo = GitRepo(self.content_dir)
        master_head = repo.run('git', 'rev-parse', 'master')
        plugin_fields = {'url': self.content_dir}
        # By default, the git plugin should reup from master.
        expected_output = {'rev': master_head}
        output = test_plugin_get_reup_fields(self.plugin_context, 'git',
                                             plugin_fields)
        self.assertDictEqual(expected_output, output)
        # Add some new commits and make sure master gets fetched properly.
        repo.run('git', 'commit', '--allow-empty', '-m', 'junk')
        repo.run('git', 'checkout', '-q', '-b', 'newbranch')
        repo.run('git', 'commit', '--allow-empty', '-m', 'more junk')
        new_master_head = repo.run('git', 'rev-parse', 'master')
        expected_output['rev'] = new_master_head
        output = test_plugin_get_reup_fields(self.plugin_context, 'git',
                                             plugin_fields)
        self.assertDictEqual(expected_output, output)
        # Now specify the reup target explicitly.
        newbranch_head = repo.run('git', 'rev-parse', 'newbranch')
        plugin_fields['reup'] = 'newbranch'
        expected_output['rev'] = newbranch_head
        output = test_plugin_get_reup_fields(self.plugin_context, 'git',
                                             plugin_fields)
        self.assertDictEqual(expected_output, output)

    @unittest.skipIf(
        sys.version_info < HG_MINIMUM_PYTHON_VERSION,
        "Python too old for hg",
    )
    def test_hg_plugin_reup(self):
        repo = HgRepo(self.content_dir)
        default_tip = repo.run('hg', 'identify', '--debug', '-r',
                               'default').split()[0]
        plugin_fields = {'url': self.content_dir}
        # By default, the hg plugin should reup from default.
        expected_output = {'rev': default_tip}
        output = test_plugin_get_reup_fields(self.plugin_context, 'hg',
                                             plugin_fields)
        self.assertDictEqual(expected_output, output)
        # Add some new commits and make sure master gets fetched properly.
        shared.write_files(self.content_dir,
                           {'randomfile': "hg doesn't like empty commits"})
        repo.run('hg', 'commit', '-A', '-m', 'junk')
        shared.write_files(
            self.content_dir,
            {'randomfile': "hg still doesn't like empty commits"})
        repo.run('hg', 'branch', 'newbranch')
        repo.run('hg', 'commit', '-A', '-m', 'more junk')
        new_default_tip = repo.run('hg', 'identify', '--debug', '-r',
                                   'default').split()[0]
        expected_output['rev'] = new_default_tip
        output = test_plugin_get_reup_fields(self.plugin_context, 'hg',
                                             plugin_fields)
        self.assertDictEqual(expected_output, output)
        # Now specify the reup target explicitly.
        newbranch_tip = repo.run('hg', 'identify', '--debug', '-r',
                                 'tip').split()[0]
        plugin_fields['reup'] = 'newbranch'
        expected_output['rev'] = newbranch_tip
        output = test_plugin_get_reup_fields(self.plugin_context, 'hg',
                                             plugin_fields)
        self.assertDictEqual(expected_output, output)

    def test_curl_plugin_fetch(self):
        curl_content = {'myfile': 'content'}
        test_dir = shared.create_dir(curl_content)
        test_url = (Path(test_dir) / 'myfile').as_uri()
        fields = {'url': test_url}
        self.do_plugin_test('curl', fields, curl_content)
        # Run the test again with an explicit hash and an explicit filename.
        digest = hashlib.sha1()
        digest.update(b'content')
        real_hash = digest.hexdigest()
        fields['sha1'] = real_hash
        fields['filename'] = 'newname'
        self.do_plugin_test('curl', fields, {'newname': 'content'})
        # Now run it with the wrong hash, and confirm that there's an error.
        fields['sha1'] = 'wrong hash'
        with self.assertRaises(plugin.PluginRuntimeError):
            self.do_plugin_test('curl', fields, {'newname': 'content'})

    def test_curl_plugin_fetch_archives(self):
        for type in 'zip', 'tar':
            fields = {
                'url': (shared.test_resources / ('with_exe.' + type)).as_uri(),
                'unpack': type,
            }
            fetch_dir = shared.create_dir()
            self.do_plugin_test(
                'curl',
                fields, {
                    'not_exe.txt': 'Not executable.\n',
                    'exe.sh': 'echo Executable.\n',
                },
                fetch_dir=fetch_dir)
            shared.assert_not_executable(
                os.path.join(fetch_dir, 'not_exe.txt'))
            shared.assert_executable(os.path.join(fetch_dir, 'exe.sh'))

    def test_curl_plugin_fetch_evil_archive(self):
        # There are several evil archives checked in under tests/resources. The
        # others are checked directly as part of test_curl_plugin.py.
        fields = {
            'url': (shared.test_resources / '.tar').as_uri(),
            'unpack': 'tar',
        }
        with self.assertRaises(plugin.PluginRuntimeError):
            self.do_plugin_test('curl', fields, {})

    def test_curl_plugin_reup(self):
        curl_content = {'myfile': 'content'}
        test_dir = shared.create_dir(curl_content)
        test_url = (Path(test_dir) / 'myfile').as_uri()
        digest = hashlib.sha1()
        digest.update(b'content')
        real_hash = digest.hexdigest()
        fields = {'url': test_url}
        output = test_plugin_get_reup_fields(self.plugin_context, 'curl',
                                             fields)
        self.assertDictEqual({'sha1': real_hash}, output)
        # Confirm that we get the same thing with a preexisting hash.
        fields['sha1'] = 'preexisting junk'
        output = test_plugin_get_reup_fields(self.plugin_context, 'curl',
                                             fields)
        self.assertDictEqual({'sha1': real_hash}, output)

    def test_cp_plugin(self):
        self.do_plugin_test("cp", {"path": self.content_dir}, self.content)

    @unittest.skipIf(os.name == 'nt', 'the rsync plugin is Unix-only')
    def test_rsync_plugin(self):
        self.do_plugin_test("rsync", {"path": self.content_dir}, self.content)

    @unittest.skipIf(os.name != 'nt', 'the bat plugin is Windows-only')
    def test_bat_plugin(self):
        self.do_plugin_test(
            "bat", {"filename": "xyz", "message": "hello Windows"},
            {"xyz": "hello Windows\n"})

    def test_empty_plugin(self):
        self.do_plugin_test("empty", {}, {})

    def test_missing_required_field(self):
        # The 'url' field is required for git.
        try:
            self.do_plugin_test('git', {}, self.content)
        except plugin.PluginModuleFieldError as e:
            assert 'url' in e.message, 'message should mention missing field'
        else:
            assert False, 'should throw PluginModuleFieldError'

    def test_unknown_field(self):
        # The 'junk' field isn't valid for git.
        bad_fields = {'url': self.content_dir, 'junk': 'junk'}
        try:
            self.do_plugin_test('git', bad_fields, self.content)
        except plugin.PluginModuleFieldError as e:
            assert 'junk' in e.message, 'message should mention bad field'
        else:
            assert False, 'should throw PluginModuleFieldError'

    def test_user_defined_plugin(self):
        plugin_prefix = 'peru/plugins/footype/'
        fetch_file = plugin_prefix + 'fetch.py'
        reup_file = plugin_prefix + 'reup.py'
        plugin_yaml_file = plugin_prefix + 'plugin.yaml'
        fake_config_dir = shared.create_dir({
            fetch_file:
            '#! /usr/bin/env python3\nprint("hey there!")\n',
            reup_file:
            textwrap.dedent('''\
                #! /usr/bin/env python3
                import os
                outfile = os.environ['PERU_REUP_OUTPUT']
                print("name: val", file=open(outfile, 'w'))
                '''),
            plugin_yaml_file:
            textwrap.dedent('''\
                sync exe: fetch.py
                reup exe: reup.py
                required fields: []
                ''')
        })
        os.chmod(os.path.join(fake_config_dir, fetch_file), 0o755)
        os.chmod(os.path.join(fake_config_dir, reup_file), 0o755)
        fetch_dir = shared.create_dir()

        # We need to trick peru into loading plugins from the fake config dir
        # dir. We do this by setting an env var, which depends on the platform.
        if os.name == 'nt':
            # Windows
            config_path_variable = 'LOCALAPPDATA'
        else:
            # non-Windows
            config_path_variable = 'XDG_CONFIG_HOME'

        with temporary_environment(config_path_variable, fake_config_dir):
            output = test_plugin_fetch(self.plugin_context, 'footype', {},
                                       fetch_dir)
            self.assertEqual('hey there!\n', output)
            output = test_plugin_get_reup_fields(self.plugin_context,
                                                 'footype', {})
            self.assertDictEqual({'name': 'val'}, output)

    def test_no_such_plugin(self):
        with self.assertRaises(plugin.PluginCandidateError):
            test_plugin_fetch(self.plugin_context, 'nosuchtype!', {},
                              os.devnull)


@contextlib.contextmanager
def temporary_environment(name, value):
    NOT_SET = object()
    old_value = os.environ.get(name, NOT_SET)
    os.environ[name] = value
    try:
        yield
    finally:
        if old_value is NOT_SET:
            del os.environ[name]
        else:
            os.environ[name] = old_value



import os
from textwrap import dedent

import shared
from shared import run_peru_command, assert_contents


class ReupIntegrationTest(shared.PeruTest):
    def setUp(self):
        self.foo_dir = shared.create_dir({'a': 'b'})
        self.foo_repo = shared.GitRepo(self.foo_dir)
        self.foo_master = self.foo_repo.run('git', 'rev-parse', 'master')
        self.bar_dir = shared.create_dir()
        self.bar_repo = shared.GitRepo(self.bar_dir)
        self.bar_repo.run('git', 'checkout', '-q', '-b', 'otherbranch')
        with open(os.path.join(self.bar_dir, 'barfile'), 'w') as f:
            f.write('new')
        self.bar_repo.run('git', 'add', '-A')
        self.bar_repo.run('git', 'commit', '-m', 'creating barfile')
        self.bar_otherbranch = self.bar_repo.run('git', 'rev-parse',
                                                 'otherbranch')

    def test_single_reup(self):
        yaml_without_imports = dedent('''\
            git module foo:
                url: {}
                rev: master

            git module bar:
                url: {}
                reup: otherbranch
            ''').format(self.foo_dir, self.bar_dir)
        test_dir = shared.create_dir({'peru.yaml': yaml_without_imports})
        expected = dedent('''\
            git module foo:
                url: {}
                rev: {}

            git module bar:
                url: {}
                reup: otherbranch
            ''').format(self.foo_dir, self.foo_master, self.bar_dir)
        run_peru_command(['reup', 'foo'], test_dir)
        assert_contents(test_dir, {'peru.yaml': expected}, excludes=['.peru'])

    def test_reup_sync(self):
        yaml_with_imports = dedent('''\
            imports:
                foo: ./
                bar: ./

            git module foo:
                url: {}
                rev: {}

            git module bar:
                url: {}
                reup: otherbranch
            ''').format(self.foo_dir, self.foo_master, self.bar_dir)
        test_dir = shared.create_dir({'peru.yaml': yaml_with_imports})
        # First reup without the sync.
        run_peru_command(['reup', 'foo', '--no-sync'], test_dir)
        assert_contents(test_dir, {}, excludes=['.peru', 'peru.yaml'])
        # Now do it with the sync. Note that barfile wasn't pulled in, because
        # we didn't reup bar.
        run_peru_command(['reup', 'foo', '--quiet'], test_dir)
        assert_contents(test_dir, {'a': 'b'}, excludes=['.peru', 'peru.yaml'])

    def test_reup_all(self):
        yaml_with_imports = dedent('''\
            imports:
                foo: ./
                bar: ./

            git module foo:
                url: {}
                rev: {}

            git module bar:
                url: {}
                reup: otherbranch
            ''').format(self.foo_dir, self.foo_master, self.bar_dir)
        test_dir = shared.create_dir({'peru.yaml': yaml_with_imports})
        expected = dedent('''\
            imports:
                foo: ./
                bar: ./

            git module foo:
                url: {}
                rev: {}

            git module bar:
                url: {}
                reup: otherbranch
                rev: {}
            ''').format(self.foo_dir, self.foo_master, self.bar_dir,
                        self.bar_otherbranch)
        run_peru_command(['reup'], test_dir)
        # This time we finally pull in barfile.
        assert_contents(
            test_dir, {
                'peru.yaml': expected,
                'a': 'b',
                'barfile': 'new'
            },
            excludes=['.peru'])



import os
from pathlib import Path

import shared


class SharedTestCodeTest(shared.PeruTest):
    def test_create_dir(self):
        empty_dir = shared.create_dir()
        self.assertListEqual([], os.listdir(empty_dir))
        content = {Path('foo'): 'a', Path('bar/baz'): 'b'}
        content_dir = shared.create_dir(content)
        # Don't use read_dir, because the read_dir test relies on create_dir.
        actual_content = {}
        for p in Path(content_dir).glob('**/*'):
            if p.is_dir():
                continue
            with p.open() as f:
                actual_content[p.relative_to(content_dir)] = f.read()
        self.assertDictEqual(content, actual_content)

    def test_read_dir(self):
        content = {Path('foo'): 'a', Path('bar/baz'): 'b'}
        test_dir = shared.create_dir(content)
        read_content = shared.read_dir(test_dir)
        self.assertDictEqual(content, read_content)
        self.assertDictEqual({
            Path('foo'): 'a'
        }, shared.read_dir(test_dir, excludes=['bar']))
        self.assertDictEqual({
            Path('foo'): 'a'
        }, shared.read_dir(test_dir, excludes=['bar/baz']))

    def test_assert_contents(self):
        content = {'foo': 'a', 'bar/baz': 'b'}
        test_dir = shared.create_dir(content)
        shared.assert_contents(test_dir, content)
        shared.write_files(test_dir, {'bing': 'c'})
        with self.assertRaises(AssertionError):
            shared.assert_contents(test_dir, content)
        shared.assert_contents(test_dir, content, excludes=['bing'])
        try:
            shared.assert_contents(test_dir, content, excludes=['foo'])
        except AssertionError as e:
            assert e.args[0].startswith('EXPECTED FILES WERE EXCLUDED')



from textwrap import dedent

import yaml

from peru import edit_yaml
import shared

yaml_template = dedent("""\
    a:
      b: [1, 2, 3]
      c: {}
    d: blarg
    """)


class EditYamlTest(shared.PeruTest):
    def test_replace(self):
        start_yaml = yaml_template.format("foo")
        new_yaml = edit_yaml.set_module_field(start_yaml, "a", "c", "bar")
        self.assertEqual(yaml_template.format("bar"), new_yaml)

    def test_insert(self):
        start_yaml = dedent("""\
            a:
               b: foo
              """)
        new_yaml = edit_yaml.set_module_field(start_yaml, "a", "c", "bar")
        self.assertEqual(start_yaml + "   c: bar\n", new_yaml)

    def test_insert_number_looking_fields(self):
        # These all need to be quoted, or else YAML will interpret them as
        # literal ints and floats.
        start_yaml = dedent('''\
            a:
              b: foo
            ''')
        intermediate = edit_yaml.set_module_field(start_yaml, 'a', 'c', '5')
        new_yaml = edit_yaml.set_module_field(intermediate, 'a', 'd', '.0')
        expected_yaml = start_yaml + '  c: "5"\n  d: ".0"\n'
        self.assertEqual(expected_yaml, new_yaml)
        self.assertDictEqual(
            yaml.safe_load(new_yaml),
            {'a': {
                'b': 'foo',
                'c': '5',
                'd': '.0',
            }})

    def test_insert_with_last_field_as_dict(self):
        start_yaml = dedent("""\
            a:
              b:
                foo: bar
                baz: bing
            x: y
            """)
        end_yaml = dedent("""\
            a:
              b:
                foo: bar
                baz: bing
              c: stuff
            x: y
            """)
        edited_yaml = edit_yaml.set_module_field(start_yaml, "a", "c", "stuff")
        self.assertEqual(end_yaml, edited_yaml)

    def test_with_file(self):
        tmp_name = shared.tmp_file()
        start_yaml = yaml_template.format("foo")
        with open(tmp_name, "w") as f:
            f.write(start_yaml)
        edit_yaml.set_module_field_in_file(tmp_name, "a", "c", "bar")
        with open(tmp_name) as f:
            new_yaml = f.read()
        self.assertEqual(yaml_template.format("bar"), new_yaml)



import os
import time
import unittest

import peru.cache
from shared import assert_contents, create_dir, make_synchronous, PeruTest, \
    COLON


class CacheTest(PeruTest):
    @make_synchronous
    async def setUp(self):
        self.cache = await peru.cache.Cache(create_dir())
        self.content = {
            'a': 'foo',
            'b/c': 'bar',
            'b/d': 'baz',
        }
        self.content_dir = create_dir(self.content)
        self.content_tree = await self.cache.import_tree(self.content_dir)

    @make_synchronous
    async def test_basic_export(self):
        export_dir = create_dir()
        await self.cache.export_tree(self.content_tree, export_dir)
        assert_contents(export_dir, self.content)

    @make_synchronous
    async def test_export_force_with_preexisting_files(self):
        # Create a working tree with a conflicting file.
        dirty_content = {'a': 'junk'}
        export_dir = create_dir(dirty_content)
        # Export should fail by default.
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            await self.cache.export_tree(self.content_tree, export_dir)
        assert_contents(export_dir, dirty_content)
        # But it should suceed with the force flag.
        await self.cache.export_tree(self.content_tree, export_dir, force=True)
        assert_contents(export_dir, self.content)

    @make_synchronous
    async def test_export_force_with_changed_files(self):
        export_dir = create_dir()
        await self.cache.export_tree(self.content_tree, export_dir)
        # If we dirty a file, a resync should fail.
        with open(os.path.join(export_dir, 'a'), 'w') as f:
            f.write('dirty')
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            await self.cache.export_tree(
                self.content_tree, export_dir, previous_tree=self.content_tree)
        # But it should succeed with the --force flag.
        await self.cache.export_tree(
            self.content_tree,
            export_dir,
            force=True,
            previous_tree=self.content_tree)
        assert_contents(export_dir, self.content)

    @make_synchronous
    async def test_multiple_imports(self):
        new_content = {'fee/fi': 'fo fum'}
        new_tree = await self.cache.import_tree(create_dir(new_content))
        export_dir = create_dir()
        await self.cache.export_tree(new_tree, export_dir)
        assert_contents(export_dir, new_content)

    @make_synchronous
    async def test_import_with_gitignore(self):
        # Make sure our git imports don't get confused by .gitignore files.
        new_content = {'fee/fi': 'fo fum', '.gitignore': 'fee/'}
        new_tree = await self.cache.import_tree(create_dir(new_content))
        export_dir = create_dir()
        await self.cache.export_tree(new_tree, export_dir)
        assert_contents(export_dir, new_content)

    @make_synchronous
    async def test_import_with_files(self):
        # Include a leading colon, to check that we escape pathspecs correctly
        # with a leading ./
        all_content = {'foo': '', 'bar': '', COLON + 'baz/bing': ''}
        test_dir = create_dir(all_content)
        tree = await self.cache.import_tree(
            test_dir, picks=['foo', COLON + 'baz'])
        expected_content = {'foo': '', COLON + 'baz/bing': ''}
        out_dir = create_dir()
        await self.cache.export_tree(tree, out_dir)
        assert_contents(out_dir, expected_content)

        # Repeat the same test with an exclude, again with a colon.
        tree = await self.cache.import_tree(
            test_dir, excludes=['foo', COLON + 'baz'])
        expected_content = {'bar': ''}
        out_dir = create_dir()
        await self.cache.export_tree(tree, out_dir)
        assert_contents(out_dir, expected_content)

    @unittest.skip(
        "broken by Git 2.34, "
        "see https://github.com/buildinspace/peru/issues/221"
    )
    @make_synchronous
    async def test_export_with_existing_files(self):
        # Create a dir with an existing file that doesn't conflict.
        more_content = {'untracked': 'stuff'}
        export_dir = create_dir(more_content)
        await self.cache.export_tree(self.content_tree, export_dir)
        expected_content = self.content.copy()
        expected_content.update(more_content)
        assert_contents(export_dir, expected_content)

        # But if we try to export twice, the export_dir will now have
        # conflicting files, and export_tree() should throw.
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            await self.cache.export_tree(self.content_tree, export_dir)

        # By default, git's checkout safety doesn't protect files that are
        # .gitignore'd. Make sure we still throw the right errors in the
        # presence of a .gitignore file.
        with open(os.path.join(export_dir, '.gitignore'), 'w') as f:
            f.write('*\n')  # .gitignore everything
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            await self.cache.export_tree(self.content_tree, export_dir)

    @make_synchronous
    async def test_previous_tree(self):
        export_dir = create_dir(self.content)

        # Create some new content.
        new_content = self.content.copy()
        new_content['a'] += ' different'
        new_content['newfile'] = 'newfile stuff'
        new_dir = create_dir(new_content)
        new_tree = await self.cache.import_tree(new_dir)

        # Now use cache.export_tree to move from the original content to the
        # different content.
        await self.cache.export_tree(
            new_tree, export_dir, previous_tree=self.content_tree)
        assert_contents(export_dir, new_content)

        # Now do the same thing again, but use a dirty working copy. This
        # should cause an error.
        dirty_content = self.content.copy()
        dirty_content['a'] += ' dirty'
        dirty_dir = create_dir(dirty_content)
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            await self.cache.export_tree(
                new_tree, dirty_dir, previous_tree=self.content_tree)
        # But if the file is simply missing, it should work.
        os.remove(os.path.join(dirty_dir, 'a'))
        await self.cache.export_tree(
            new_tree, dirty_dir, previous_tree=self.content_tree)
        assert_contents(dirty_dir, new_content)

        # Make sure we get an error even if the dirty file is unchanged between
        # the previous tree and the new one.
        no_conflict_dirty_content = self.content.copy()
        no_conflict_dirty_content['b/c'] += ' dirty'
        no_conflict_dirty_dir = create_dir(no_conflict_dirty_content)
        with self.assertRaises(peru.cache.DirtyWorkingCopyError):
            await self.cache.export_tree(
                new_tree,
                no_conflict_dirty_dir,
                previous_tree=self.content_tree)

    @make_synchronous
    async def test_missing_files_in_previous_tree(self):
        '''Export should allow missing files, and it should recreate them.'''
        export_dir = create_dir()
        # Nothing in content_tree exists yet, so this export should be the same
        # as if previous_tree wasn't specified.
        await self.cache.export_tree(
            self.content_tree, export_dir, previous_tree=self.content_tree)
        assert_contents(export_dir, self.content)
        # Make sure the same applies with just a single missing file.
        os.remove(os.path.join(export_dir, 'a'))
        await self.cache.export_tree(
            self.content_tree, export_dir, previous_tree=self.content_tree)
        assert_contents(export_dir, self.content)
        # This should work even if we run it from a subdirectory of
        # content_tree. (Past buggy behavior here:
        # https://github.com/buildinspace/peru/issues/210.)
        prev_dir = os.getcwd()
        try:
            os.remove(os.path.join(export_dir, 'a'))
            os.chdir(os.path.join(export_dir, 'b'))
            await self.cache.export_tree(
                self.content_tree, export_dir, previous_tree=self.content_tree)
            assert_contents(export_dir, self.content)
        finally:
            os.chdir(prev_dir)

    @make_synchronous
    async def test_merge_trees(self):
        merged_tree = await self.cache.merge_trees(self.content_tree,
                                                   self.content_tree, 'subdir')
        expected_content = dict(self.content)
        for path, content in self.content.items():
            expected_content[os.path.join('subdir', path)] = content
        export_dir = create_dir()
        await self.cache.export_tree(merged_tree, export_dir)
        assert_contents(export_dir, expected_content)

        with self.assertRaises(peru.cache.MergeConflictError):
            # subdir/ is already populated, so this merge should throw.
            await self.cache.merge_trees(merged_tree, self.content_tree,
                                         'subdir')

    @make_synchronous
    async def test_merge_with_deep_prefix(self):
        '''This test was inspired by a bug on Windows where we would give git a
        backslash-separated merge prefix, even though git demands forward slash
        as a path separator.'''
        content = {'file': 'stuff'}
        content_dir = create_dir(content)
        tree = await self.cache.import_tree(content_dir)
        prefixed_tree = await self.cache.merge_trees(None, tree, 'a/b/')
        export_dir = create_dir()
        await self.cache.export_tree(prefixed_tree, export_dir)
        assert_contents(export_dir, {'a/b/file': 'stuff'})

    @make_synchronous
    async def test_read_file(self):
        # Include a leading colon, to check that we escape pathspecs correctly
        # with a leading ./
        all_content = {'a': 'foo', COLON + 'b/c': 'bar'}
        test_dir = create_dir(all_content)
        tree = await self.cache.import_tree(test_dir)
        a_content = await self.cache.read_file(tree, 'a')
        bc_content = await self.cache.read_file(tree, COLON + 'b/c')
        self.assertEqual(b'foo', a_content)
        self.assertEqual(b'bar', bc_content)
        with self.assertRaises(FileNotFoundError):
            await self.cache.read_file(tree, 'nonexistent')
        with self.assertRaises(IsADirectoryError):
            await self.cache.read_file(tree, COLON + 'b')

    # A helper method for several tests below below.
    async def do_excludes_and_files_test(self, excludes, picks, expected):
        tree = await self.cache.import_tree(
            self.content_dir, excludes=excludes, picks=picks)
        out_dir = create_dir()
        await self.cache.export_tree(tree, out_dir)
        assert_contents(out_dir, expected)

    @make_synchronous
    async def test_import_with_specific_file(self):
        await self.do_excludes_and_files_test(
            excludes=[], picks=['a'], expected={'a': 'foo'})

    @make_synchronous
    async def test_import_with_specific_dir(self):
        await self.do_excludes_and_files_test(
            excludes=[], picks=['b'], expected={
                'b/c': 'bar',
                'b/d': 'baz'
            })

    @make_synchronous
    async def test_import_with_excluded_file(self):
        await self.do_excludes_and_files_test(
            excludes=['a'], picks=[], expected={
                'b/c': 'bar',
                'b/d': 'baz'
            })

    @make_synchronous
    async def test_import_with_excluded_dir(self):
        await self.do_excludes_and_files_test(
            excludes=['b'], picks=[], expected={'a': 'foo'})

    @make_synchronous
    async def test_import_with_excludes_and_files(self):
        await self.do_excludes_and_files_test(
            excludes=['b/c'], picks=['b'], expected={'b/d': 'baz'})

    @make_synchronous
    async def test_ls_tree(self):
        # Use the recursive case to get valid entries for each file. We could
        # hardcode these, but it would be messy and annoying to maintain.
        entries = await self.cache.ls_tree(self.content_tree, recursive=True)
        assert entries.keys() == {'a', 'b', 'b/c', 'b/d'}
        assert (entries['a'].type == entries['b/c'].type == entries['b/d'].type
                == peru.cache.BLOB_TYPE)
        assert entries['b'].type == peru.cache.TREE_TYPE

        # Check the non-recursive, non-path case.
        self.assertDictEqual({
            'a': entries['a'],
            'b': entries['b']
        }, (await self.cache.ls_tree(self.content_tree)))

        # Check the single file case, and make sure paths are normalized.
        self.assertDictEqual({
            'b/c': entries['b/c']
        }, (await self.cache.ls_tree(self.content_tree, 'b/c//./')))

        # Check the single dir case. (Trailing slash shouldn't matter, because
        # we nomalize it, but git will do the wrong thing if we forget
        # normalization.)
        self.assertDictEqual({
            'b': entries['b']
        }, (await self.cache.ls_tree(self.content_tree, 'b/')))

        # Check the recursive dir case.
        self.assertDictEqual({
            'b': entries['b'],
            'b/c': entries['b/c'],
            'b/d': entries['b/d']
        }, (await self.cache.ls_tree(self.content_tree, 'b', recursive=True)))

        # Make sure that we don't skip over a target file in recursive mode.
        self.assertDictEqual({
            'b/c': entries['b/c']
        }, (await self.cache.ls_tree(self.content_tree, 'b/c',
                                     recursive=True)))

    @make_synchronous
    async def test_modify_tree(self):
        base_dir = create_dir({'a': 'foo', 'b/c': 'bar'})
        base_tree = await self.cache.import_tree(base_dir)
        entries = await self.cache.ls_tree(base_tree, recursive=True)
        cases = []

        # Test regular deletions.
        cases.append(({'a': None}, {'b/c': 'bar'}))
        cases.append((
            {
                'a//./': None
            },  # Paths should get normalized.
            {
                'b/c': 'bar'
            }))
        cases.append(({'b': None}, {'a': 'foo'}))
        cases.append(({'b/c': None}, {'a': 'foo'}))
        cases.append(({'x/y/z': None}, {'a': 'foo', 'b/c': 'bar'}))
        cases.append(({'b/x': None}, {'a': 'foo', 'b/c': 'bar'}))
        # Test the case where we try to delete below a file.
        cases.append(({'a/x': None}, {'a': 'foo', 'b/c': 'bar'}))
        # Test insertions.
        cases.append(({'b': entries['a']}, {'a': 'foo', 'b': 'foo'}))
        cases.append(({
            'x': entries['a']
        }, {
            'a': 'foo',
            'x': 'foo',
            'b/c': 'bar'
        }))
        cases.append(({
            'x': entries['b']
        }, {
            'a': 'foo',
            'b/c': 'bar',
            'x/c': 'bar'
        }))
        cases.append(({
            'd/e/f': entries['a']
        }, {
            'a': 'foo',
            'b/c': 'bar',
            'd/e/f': 'foo'
        }))
        cases.append(({
            'd/e/f': entries['b']
        }, {
            'a': 'foo',
            'b/c': 'bar',
            'd/e/f/c': 'bar'
        }))

        for modifications, result in cases:
            modified_tree = await self.cache.modify_tree(
                base_tree, modifications)
            modified_dir = create_dir()
            await self.cache.export_tree(modified_tree, modified_dir)
            error_msg = ('modify_tree failed to give result {} '
                         'for modifications {}'.format(
                             repr(result), repr(modifications)))
            assert_contents(modified_dir, result, message=error_msg)

    @make_synchronous
    async def test_git_attributes(self):
        # Setting the 'text' attribute when files contain Windows-style
        # newlines makes them appear dirty, which leads to errors where the
        # cache thinks its own checked out files are dirty. (I don't honestly
        # understand all the details.) The cache's git calls will read
        # .gitattributes in the sync dir, so we need to set our own attributes
        # in the $GIT_DIR to override. Everything in this test has to be done
        # in binary mode or it will all get muddled up when we actually run it
        # on Windows.
        windows_content = {'file': b'windows newline\r\n'}
        gitattributes_content = {'.gitattributes': b'* text'}
        both_content = windows_content.copy()
        both_content.update(gitattributes_content)
        windows_dir = create_dir(windows_content)
        tree = await self.cache.import_tree(windows_dir)
        out_dir = create_dir(gitattributes_content)
        # This export fails without the fix mentioned above.
        await self.cache.export_tree(tree, out_dir)
        assert_contents(out_dir, both_content, binary=True)

    @make_synchronous
    async def test_touched_file(self):
        # Bumping the mtime on a file makes it appear dirty to `git
        # diff-files`. However, when the index is refreshed with `git
        # update-index`, the dirtiness should go away. This test guarantees
        # that we do that refresh, both with and without a cached index file.
        # Note that because the index file only has an mtime resolution of 1
        # second, we have to artificially inflate the mtime to guarantee that
        # the file will actually appear dirty.
        export_dir = create_dir()
        a_path = os.path.join(export_dir, 'a')
        t = time.time()

        def bump_mtime_one_minute():
            nonlocal t
            t += 60  # Add a whole minute to the mtime we set.
            os.utime(a_path, (t, t))

        # Do the first export.
        await self.cache.export_tree(self.content_tree, export_dir)
        # Touch a and rerun the export with no cached index.
        bump_mtime_one_minute()
        await self.cache.export_tree(
            self.content_tree, export_dir, previous_tree=self.content_tree)
        # Create a cached index file.
        index_dir = create_dir()
        index_file = os.path.join(index_dir, 'test_index_file')
        await self.cache.export_tree(
            self.content_tree,
            export_dir,
            previous_tree=self.content_tree,
            previous_index_file=index_file)
        # Finally, touch a again and rerun the export using the cached index.
        bump_mtime_one_minute()
        await self.cache.export_tree(
            self.content_tree,
            export_dir,
            previous_tree=self.content_tree,
            previous_index_file=index_file)

    @make_synchronous
    async def test_import_ignores_dotperu(self):
        # We have a security problem similar to git's if we allow '.peru'
        # directories in the trees we write to disk. (See
        # https://github.com/blog/1938-vulnerability-announced-update-your-git-clients.)
        # We need to check that *all* '.peru' dirs are ignored in imported
        # trees, including inside of nested subdirectories. And as in the git
        # issue, we need to do this in a case-insensitive way.
        content = {
            'foo': 'bar',
            '.peru/foo1': 'bar',
            '.PERU/foo2': 'bar',
            '.pErU/foo3': 'bar',
            'dir/foo': 'bar',
            'dir/.peru/foo1': 'bar',
            'dir/.PERU/foo2': 'bar',
            'dir/.peRU/foo3': 'bar',
        }
        tree = await self.cache.import_tree(create_dir(content))
        entries = await self.cache.ls_tree(tree, recursive=True)
        self.assertEqual({'foo', 'dir', 'dir/foo'}, entries.keys(),
                         "Expected all of the .peru dirs to be omitted.")

    @make_synchronous
    async def test_validate_capitalizations(self):
        # Assert that the set of capitalizations is the right size, both before
        # and after deduplication.
        self.assertEqual(len(peru.cache.DOTPERU_CAPITALIZATIONS), 16)
        self.assertEqual(len(set(peru.cache.DOTPERU_CAPITALIZATIONS)), 16)
        # Check that every capitalization actually spells ".peru".
        for capitalization in peru.cache.DOTPERU_CAPITALIZATIONS:
            self.assertEqual(capitalization.lower(), ".peru")



# This definition of AsyncExitStack is copied entirely from
# https://github.com/sorcio/async_exit_stack, in order to support Python 3.5
# and 3.6.
#
# flake8: noqa

from collections import deque
import sys
from types import MethodType


class AsyncExitStack:
    """Async context manager for dynamic management of a stack of exit
    callbacks.
    For example:
        async with AsyncExitStack() as stack:
            connections = [await stack.enter_async_context(get_connection())
                for i in range(5)]
            # All opened connections will automatically be released at the
            # end of the async with statement, even if attempts to open a
            # connection later in the list raise an exception.
    """

    ### _BaseExitStack staticmethods

    @staticmethod
    def _create_exit_wrapper(cm, cm_exit):
        return MethodType(cm_exit, cm)

    @staticmethod
    def _create_cb_wrapper(callback, *args, **kwds):
        def _exit_wrapper(exc_type, exc, tb):
            callback(*args, **kwds)

        return _exit_wrapper

    ### AsyncExitStack staticmethods

    @staticmethod
    def _create_async_exit_wrapper(cm, cm_exit):
        return MethodType(cm_exit, cm)

    @staticmethod
    def _create_async_cb_wrapper(callback, *args, **kwds):
        async def _exit_wrapper(exc_type, exc, tb):
            await callback(*args, **kwds)

        return _exit_wrapper

    ### _BaseExitStack methods

    def __init__(self):
        self._exit_callbacks = deque()

    def pop_all(self):
        """Preserve the context stack by transferring it to a new instance."""
        new_stack = type(self)()
        new_stack._exit_callbacks = self._exit_callbacks
        self._exit_callbacks = deque()
        return new_stack

    def push(self, exit):
        """Registers a callback with the standard __exit__ method signature.
        Can suppress exceptions the same way __exit__ method can.
        Also accepts any object with an __exit__ method (registering a call
        to the method instead of the object itself).
        """
        # We use an unbound method rather than a bound method to follow
        # the standard lookup behaviour for special methods.
        _cb_type = type(exit)

        try:
            exit_method = _cb_type.__exit__
        except AttributeError:
            # Not a context manager, so assume it's a callable.
            self._push_exit_callback(exit)
        else:
            self._push_cm_exit(exit, exit_method)
        return exit  # Allow use as a decorator.

    def enter_context(self, cm):
        """Enters the supplied context manager.
        If successful, also pushes its __exit__ method as a callback and
        returns the result of the __enter__ method.
        """
        # We look up the special methods on the type to match the with
        # statement.
        _cm_type = type(cm)
        _exit = _cm_type.__exit__
        result = _cm_type.__enter__(cm)
        self._push_cm_exit(cm, _exit)
        return result

    def callback(self, callback, *args, **kwds):
        """Registers an arbitrary callback and arguments.
        Cannot suppress exceptions.
        """
        _exit_wrapper = self._create_cb_wrapper(callback, *args, **kwds)

        # We changed the signature, so using @wraps is not appropriate, but
        # setting __wrapped__ may still help with introspection.
        _exit_wrapper.__wrapped__ = callback
        self._push_exit_callback(_exit_wrapper)
        return callback  # Allow use as a decorator

    def _push_cm_exit(self, cm, cm_exit):
        """Helper to correctly register callbacks to __exit__ methods."""
        _exit_wrapper = self._create_exit_wrapper(cm, cm_exit)
        self._push_exit_callback(_exit_wrapper, True)

    def _push_exit_callback(self, callback, is_sync=True):
        self._exit_callbacks.append((is_sync, callback))

    ### AsyncExitStack methods

    async def enter_async_context(self, cm):
        """Enters the supplied async context manager.
        If successful, also pushes its __aexit__ method as a callback and
        returns the result of the __aenter__ method.
        """
        _cm_type = type(cm)
        _exit = _cm_type.__aexit__
        result = await _cm_type.__aenter__(cm)
        self._push_async_cm_exit(cm, _exit)
        return result

    def push_async_exit(self, exit):
        """Registers a coroutine function with the standard __aexit__ method
        signature.
        Can suppress exceptions the same way __aexit__ method can.
        Also accepts any object with an __aexit__ method (registering a call
        to the method instead of the object itself).
        """
        _cb_type = type(exit)
        try:
            exit_method = _cb_type.__aexit__
        except AttributeError:
            # Not an async context manager, so assume it's a coroutine function
            self._push_exit_callback(exit, False)
        else:
            self._push_async_cm_exit(exit, exit_method)
        return exit  # Allow use as a decorator

    def push_async_callback(self, callback, *args, **kwds):
        """Registers an arbitrary coroutine function and arguments.
        Cannot suppress exceptions.
        """
        _exit_wrapper = self._create_async_cb_wrapper(callback, *args, **kwds)

        # We changed the signature, so using @wraps is not appropriate, but
        # setting __wrapped__ may still help with introspection.
        _exit_wrapper.__wrapped__ = callback
        self._push_exit_callback(_exit_wrapper, False)
        return callback  # Allow use as a decorator

    async def aclose(self):
        """Immediately unwind the context stack."""
        await self.__aexit__(None, None, None)

    def _push_async_cm_exit(self, cm, cm_exit):
        """Helper to correctly register coroutine function to __aexit__
        method."""
        _exit_wrapper = self._create_async_exit_wrapper(cm, cm_exit)
        self._push_exit_callback(_exit_wrapper, False)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc_details):
        received_exc = exc_details[0] is not None

        # We manipulate the exception state so it behaves as though
        # we were actually nesting multiple with statements
        frame_exc = sys.exc_info()[1]

        def _fix_exception_context(new_exc, old_exc):
            # Context may not be correct, so find the end of the chain
            while 1:
                exc_context = new_exc.__context__
                if exc_context is old_exc:
                    # Context is already set correctly (see issue 20317)
                    return
                if exc_context is None or exc_context is frame_exc:
                    break
                new_exc = exc_context
            # Change the end of the chain to point to the exception
            # we expect it to reference
            new_exc.__context__ = old_exc

        # Callbacks are invoked in LIFO order to match the behaviour of
        # nested context managers
        suppressed_exc = False
        pending_raise = False
        while self._exit_callbacks:
            is_sync, cb = self._exit_callbacks.pop()
            try:
                if is_sync:
                    cb_suppress = cb(*exc_details)
                else:
                    cb_suppress = await cb(*exc_details)

                if cb_suppress:
                    suppressed_exc = True
                    pending_raise = False
                    exc_details = (None, None, None)
            except:
                new_exc_details = sys.exc_info()
                # simulate the stack of exceptions by setting the context
                _fix_exception_context(new_exc_details[1], exc_details[1])
                pending_raise = True
                exc_details = new_exc_details
        if pending_raise:
            try:
                # bare "raise exc_details[1]" replaces our carefully
                # set-up context
                fixed_ctx = exc_details[1].__context__
                raise exc_details[1]
            except BaseException:
                exc_details[1].__context__ = fixed_ctx
                raise
        return received_exc and suppressed_exc



from peru.async_helpers import run_task
import peru.scope
import shared


class ScopeTest(shared.PeruTest):
    def test_parse_target(self):
        scope = scope_tree_to_scope({
            'modules': {
                'a': {
                    'modules': {
                        'b': {
                            'modules': {
                                'c': {}
                            },
                            'rules': ['r'],
                        }
                    }
                }
            }
        })
        c, (r, ) = run_task(scope.parse_target(DummyRuntime(), 'a.b.c|a.b.r'))
        assert type(c) is DummyModule and c.name == 'a.b.c'
        assert type(r) is DummyRule and r.name == 'a.b.r'


def scope_tree_to_scope(tree, prefix=""):
    '''This function is for generating dummy scope/module/rule hierarchies for
    testing. A scope tree contains a modules dictionary and a rules list, both
    optional. The values of the modules dictionary are themselves scope trees.
    So if module A contains module B and rule R, that's represented as:

    {
        'modules': {
            'A': {
                'modules': {
                    'B': {},
                },
                'rules': ['R'],
            }
        }
    }
    '''
    modules = {}
    if 'modules' in tree:
        for module_name, sub_tree in tree['modules'].items():
            full_name = prefix + module_name
            new_prefix = full_name + peru.scope.SCOPE_SEPARATOR
            module_scope = scope_tree_to_scope(sub_tree, new_prefix)
            modules[module_name] = DummyModule(full_name, module_scope)
    rules = {}
    if 'rules' in tree:
        for rule_name in tree['rules']:
            full_name = prefix + rule_name
            rules[rule_name] = DummyRule(full_name)
    return peru.scope.Scope(modules, rules)


class DummyModule:
    def __init__(self, name, scope):
        self.name = name
        self.scope = scope

    async def parse_peru_file(self, dummy_runtime):
        return self.scope, None


class DummyRule:
    def __init__(self, name):
        self.name = name


class DummyRuntime:
    pass



import collections
import re

import peru.glob as glob
import shared


class GlobTest(shared.PeruTest):
    def test_split_on_stars_interpreting_backslashes(self):
        cases = [
            ('', ['']),
            ('*', ['', '']),
            ('abc', ['abc']),
            ('abc\\', ['abc\\']),
            ('abc\\n', ['abc\\n']),
            ('abc\\\\', ['abc\\']),
            ('ab*c', ['ab', 'c']),
            ('*abc*', ['', 'abc', '']),
            (r'a\*bc', ['a*bc']),
            (r'a\\*bc', ['a\\', 'bc']),
            (r'a\\\*bc', ['a\\*bc']),
            (r'a\\\\*bc', ['a\\\\', 'bc']),
        ]
        for input, output in cases:
            self.assertEqual(
                output, glob.split_on_stars_interpreting_backslashes(input),
                'Failed split for input {}'.format(input))

    def test_glob_to_path_regex(self):
        Case = collections.namedtuple('Case', ['glob', 'matches', 'excludes'])
        cases = [
            Case(
                glob='a/b/c',
                matches=['a/b/c'],
                excludes=['a/b', 'a/b/c/', '/a/b/c', 'a/b/c/d']),
            # * should be able to match nothing.
            Case(
                glob='a/*b/c',
                matches=['a/b/c', 'a/xb/c'],
                excludes=['a/x/c', 'a/c', 'a//c']),
            # But * by itself should never match an empty path component.
            Case(
                glob='a/*/c',
                matches=['a/b/c', 'a/boooo/c', 'a/*/c'],
                excludes=['a/c', 'a/b/d/c', 'a//c']),
            # Similarly, ** does not match empty path components. It's tempting
            # to allow this, but we never want '**/c' to match '/c'.
            Case(
                glob='a/**/c',
                matches=['a/b/c', 'a/d/e/f/g/c', 'a/c'],
                excludes=['a/b/c/d', 'x/a/c', 'a//c']),
            Case(
                glob='a/**/**/c',
                matches=['a/b/c', 'a/d/e/f/g/c', 'a/c'],
                excludes=['a/b/c/d', 'x/a/c', 'a//c']),
            Case(glob='**/c', matches=['a/b/c', 'c'], excludes=['/c', 'c/d']),
            Case(
                glob='**/*/c', matches=['a/b/c', 'a/c'], excludes=['c', '/c']),
            # Leading slashes should be preserved if present.
            Case(glob='/a', matches=['/a'], excludes=['a']),
            Case(
                glob='/**/c',
                matches=['/a/b/c', '/c'],
                excludes=['c', 'a/b/c']),
            # Make sure special characters are escaped properly.
            Case(glob='a|b', matches=['a|b'], excludes=['a', 'b']),
            # Test escaped * characters.
            Case(glob='a\\*', matches=['a*'], excludes=['a', 'aa']),
        ]
        for case in cases:
            regex = glob.glob_to_path_regex(case.glob)
            for m in case.matches:
                assert re.match(regex, m), \
                    'Glob {} (regex: {} ) should match path {}'.format(
                        case.glob, regex, m)
            for e in case.excludes:
                assert not re.match(regex, e), \
                    'Glob {} (regex: {} ) should not match path {}'.format(
                    case.glob, regex, e)

    def test_bad_globs(self):
        bad_globs = [
            '**',
            'a/b/**',
            'a/b/**/',
            'a/b/**c/d',
        ]
        for bad_glob in bad_globs:
            with self.assertRaises(glob.GlobError):
                glob.glob_to_path_regex(bad_glob)

    def test_unglobbed_prefix(self):
        assert glob.unglobbed_prefix('a/b/c*/d') == 'a/b'
        assert glob.unglobbed_prefix('a/b/**/d') == 'a/b'
        assert glob.unglobbed_prefix('/a/b/*/d') == '/a/b'
        assert glob.unglobbed_prefix('*/a/b') == ''



import shared

from peru.keyval import KeyVal


class KeyValTest(shared.PeruTest):
    def test_keyval(self):
        root = shared.create_dir()
        tmp_dir = shared.create_dir()
        keyval = KeyVal(root, tmp_dir)
        key = "mykey"
        # keyval should be empty
        self.assertFalse(key in keyval)
        self.assertSetEqual(set(keyval), set())
        # set a key
        keyval[key] = "myval"
        self.assertEqual(keyval[key], "myval")
        self.assertTrue(key in keyval)
        self.assertSetEqual(set(keyval), {key})
        # overwrite the value
        keyval[key] = "anotherval"
        self.assertEqual(keyval[key], "anotherval")
        # instantiate a second keyval on the same dir, should have same content
        another_keyval = KeyVal(root, tmp_dir)
        self.assertTrue(key in another_keyval)
        self.assertEqual(another_keyval[key], "anotherval")
        self.assertSetEqual(set(another_keyval), {key})
        # test deletions
        del keyval[key]
        self.assertFalse(key in keyval)
        self.assertFalse(key in another_keyval)



#! /usr/bin/env python3

from collections import namedtuple
import configparser
import hashlib
import os
import subprocess
import sys

Result = namedtuple("Result", ["returncode", "output"])


def git(*args, git_dir=None, capture_output=False, checked=True):
    # Avoid forgetting this arg.
    assert git_dir is None or os.path.isdir(git_dir)

    command = ['git']
    if git_dir:
        command.append('--git-dir={0}'.format(git_dir))
    command.extend(args)

    stdout = subprocess.PIPE if capture_output else None
    # Always let stderr print to the caller.
    process = subprocess.Popen(
        command,
        stdin=subprocess.DEVNULL,
        stdout=stdout,
        universal_newlines=True)
    output, _ = process.communicate()
    if checked and process.returncode != 0:
        sys.exit(1)

    return Result(process.returncode, output)


def has_clone(url):
    return os.path.exists(repo_cache_path(url))


def clone_if_needed(url):
    repo_path = repo_cache_path(url)
    if not has_clone(url):
        # We look for this print in test, to count the number of clones we did.
        print('git clone ' + url)
        git('clone', '--mirror', '--progress', url, repo_path)
    return repo_path


def repo_cache_path(url):
    # Because peru gives each plugin a unique cache dir based on its cacheable
    # fields (in this case, url) we could clone directly into cache_root.
    # However, because the git plugin needs to handle git submodules as well,
    # it still has to separate things out by repo url.
    CACHE_ROOT = os.environ['PERU_PLUGIN_CACHE']

    # If we just concatenate the escaped repo URL into the path, we start to
    # run up against the 260-character path limit on Windows.
    url_hash = hashlib.sha1(url.encode()).hexdigest()

    return os.path.join(CACHE_ROOT, url_hash)


def git_fetch(url, repo_path):
    print('git fetch ' + url)
    git('fetch', '--prune', git_dir=repo_path)


def already_has_rev(repo, rev):
    # Make sure the rev exists.
    cat_result = git('cat-file', '-e', rev, git_dir=repo, checked=False)
    if cat_result.returncode != 0:
        return False
    # Get the hash for the rev.
    parse_result = git(
        'rev-parse', rev, git_dir=repo, checked=False, capture_output=True)
    if parse_result.returncode != 0:
        return False
    # Only return True for revs that are absolute hashes.
    # We could consider treating tags the way, but...
    # 1) Tags actually can change.
    # 2) It's not clear at a glance if something is a branch or a tag.
    # Keep it simple.
    return parse_result.output.strip() == rev


def checkout_tree(url, rev, dest):
    repo_path = clone_if_needed(url)
    if not already_has_rev(repo_path, rev):
        git_fetch(url, repo_path)
    # If we just use `git checkout rev -- .` here, we get an error when rev is
    # an empty commit.
    git('--work-tree=' + dest, 'read-tree', rev, git_dir=repo_path)
    git('--work-tree=' + dest, 'checkout-index', '--all', git_dir=repo_path)
    checkout_submodules(url, repo_path, rev, dest)


def checkout_submodules(parent_url, repo_path, rev, work_tree):
    if os.environ['PERU_MODULE_SUBMODULES'] == 'false':
        return

    gitmodules = os.path.join(work_tree, '.gitmodules')
    if not os.path.exists(gitmodules):
        return

    parser = configparser.ConfigParser()
    parser.read(gitmodules)
    for section in parser.sections():
        sub_relative_path = parser[section]['path']
        sub_full_path = os.path.join(work_tree, sub_relative_path)
        raw_sub_url = parser[section]['url']
        # Submodules can begin with ./ or ../, in which case they're relative
        # to the parent's URL. Handle this case.
        sub_url = expand_relative_submodule_url(raw_sub_url, parent_url)
        ls_tree = git(
            'ls-tree',
            rev,
            sub_relative_path,
            git_dir=repo_path,
            capture_output=True).output
        # Normally when you run `git submodule add ...`, git puts two things in
        # your repo: an entry in .gitmodules, and a commit object at the
        # appropriate path inside your repo. However, it's possible for those
        # two to get out of sync, especially if you use mv/rm on a directory
        # followed by `git add`, instead of the smarter `git mv`/`git rm`. If
        # we run into one of these missing submodules, just skip it.
        if len(ls_tree.strip()) == 0:
            print('WARNING: submodule ' + sub_relative_path +
                  ' is configured in .gitmodules, but missing in the repo')
            continue
        sub_rev = ls_tree.split()[2]
        checkout_tree(sub_url, sub_rev, sub_full_path)


# According to comments in its own source code, git's implementation of
# relative submodule URLs is full of unintended corner cases. See:
# https://github.com/git/git/blob/v2.20.1/builtin/submodule--helper.c#L135
#
# We absolutely give up on trying to replicate their logic -- which probably
# isn't stable in any case -- and instead we just leave the dots in and let the
# host make sense of it. A quick sanity check on GitHub confirmed that that
# seems to work for now.
def expand_relative_submodule_url(raw_sub_url, parent_url):
    if not raw_sub_url.startswith("./") and not raw_sub_url.startswith("../"):
        return raw_sub_url
    new_path = parent_url
    if not new_path.endswith("/"):
        new_path += "/"
    new_path += raw_sub_url
    return new_path


def plugin_sync(url, rev):
    checkout_tree(url, rev, os.environ['PERU_SYNC_DEST'])


def plugin_reup(url, reup):
    reup_output = os.environ['PERU_REUP_OUTPUT']
    repo_path = clone_if_needed(url)
    git_fetch(url, repo_path)
    output = git(
        'rev-parse', reup, git_dir=repo_path, capture_output=True).output
    with open(reup_output, 'w') as out_file:
        print('rev:', output.strip(), file=out_file)


def git_default_branch(url) -> str:
    """
    This function checks if the default branch is master.
    If it is not found, then it assumes it is main.
    For other default branches, user should use the 'rev' option.

    Args:
        url (str): url from the target repository to be checked.
    Returns:
        str: returns a possible match for the git default branch.
    """
    repo_path = clone_if_needed(url)
    output = git('show-ref', '--verify', '--quiet', 'refs/heads/master',
                 git_dir=repo_path, checked=False, capture_output=True)
    if output.returncode == 0:
        return 'master'
    else:
        return 'main'


def main():
    URL = os.environ['PERU_MODULE_URL']
    default_branch = git_default_branch(URL)
    REV = os.environ['PERU_MODULE_REV'] or default_branch
    REUP = os.environ['PERU_MODULE_REUP'] or default_branch

    command = os.environ['PERU_PLUGIN_COMMAND']
    if command == 'sync':
        plugin_sync(URL, REV)
    elif command == 'reup':
        plugin_reup(URL, REUP)
    else:
        raise RuntimeError('Unknown command: ' + repr(command))


if __name__ == "__main__":
    main()



from textwrap import dedent

from peru import parser
from peru.parser import parse_string, ParserError
from peru.module import Module
from peru.rule import Rule
import shared


class ParserTest(shared.PeruTest):
    def test_parse_empty_file(self):
        scope, imports = parse_string('')
        self.assertDictEqual(scope.modules, {})
        self.assertDictEqual(scope.rules, {})
        self.assertEqual(imports, {})

    def test_parse_rule(self):
        input = dedent("""\
            rule foo:
                export: out/
            """)
        scope, imports = parse_string(input)
        self.assertIn("foo", scope.rules)
        rule = scope.rules["foo"]
        self.assertIsInstance(rule, Rule)
        self.assertEqual(rule.name, "foo")
        self.assertEqual(rule.export, "out/")

    def test_parse_module(self):
        input = dedent("""\
            sometype module foo:
                url: http://www.example.com/
                rev: abcdefg
            """)
        scope, imports = parse_string(input)
        self.assertIn("foo", scope.modules)
        module = scope.modules["foo"]
        self.assertIsInstance(module, Module)
        self.assertEqual(module.name, "foo")
        self.assertEqual(module.type, "sometype")
        self.assertDictEqual(module.plugin_fields, {
            "url": "http://www.example.com/",
            "rev": "abcdefg"
        })

    def test_parse_module_default_rule(self):
        input = dedent("""\
            git module bar:
                export: bar
            """)
        scope, imports = parse_string(input)
        self.assertIn("bar", scope.modules)
        module = scope.modules["bar"]
        self.assertIsInstance(module, Module)
        self.assertIsInstance(module.default_rule, Rule)
        self.assertEqual(module.default_rule.export, "bar")

    def test_parse_toplevel_imports(self):
        input = dedent("""\
            imports:
                foo: bar/
            """)
        scope, imports = parse_string(input)
        self.assertDictEqual(scope.modules, {})
        self.assertDictEqual(scope.rules, {})
        self.assertEqual(imports, {'foo': ('bar/', )})

    def test_parse_multimap_imports(self):
        input = dedent('''\
            imports:
                foo:
                  - bar/
            ''')
        scope, imports = parse_string(input)
        self.assertDictEqual(scope.modules, {})
        self.assertDictEqual(scope.rules, {})
        self.assertEqual(imports, {'foo': ('bar/', )})

    def test_parse_empty_imports(self):
        input = dedent('''\
            imports:
            ''')
        scope, imports = parse_string(input)
        self.assertDictEqual(scope.modules, {})
        self.assertDictEqual(scope.rules, {})
        self.assertEqual(imports, {})

    def test_parse_wrong_type_imports_throw(self):
        with self.assertRaises(ParserError):
            parse_string('imports: 5')

    def test_bad_toplevel_field_throw(self):
        with self.assertRaises(ParserError):
            parse_string("foo: bar")

    def test_bad_rule_field_throw(self):
        with self.assertRaises(ParserError):
            parse_string(
                dedent("""\
                rule foo:
                    bad_field: junk
                """))

    def test_bad_rule_name_throw(self):
        with self.assertRaises(ParserError):
            parse_string("rule foo bar:")

    def test_bad_module_name_throw(self):
        with self.assertRaises(ParserError):
            parse_string("git module abc def:")
        with self.assertRaises(ParserError):
            parse_string("git module:")

    def test_duplicate_names_throw(self):
        # Modules and rules should not conflict.
        ok_input = dedent('''
            rule foo:
            git module foo:
            ''')
        parse_string(ok_input)
        # But duplicate modules should fail. (Duplicate rules are a not
        # currently possible, because their YAML keys would be exact
        # duplicates.)
        bad_input = dedent('''
            git module foo:
            hg module foo:
            ''')
        with self.assertRaises(ParserError):
            parse_string(bad_input)

    def test_non_string_module_field_name(self):
        input = dedent('''\
            git module foo:
                12345: bar
            ''')
        try:
            parse_string(input)
        except ParserError as e:
            assert '12345' in e.message
        else:
            assert False, 'expected ParserError'

    def test_non_string_module_field_value(self):
        input = dedent('''\
            git module foo:
                bar: 123
                # These booleans should turn into "true" and "false".
                baz: yes
                bing: no
            ''')
        scope, imports = parse_string(input)
        foo = scope.modules['foo']
        self.assertDictEqual(foo.plugin_fields, {
            "bar": "123",
            "baz": "true",
            "bing": "false",
        })

    def test_build_field_deprecated_message(self):
        input = dedent('''\
            rule foo:
                build: shell command
            ''')
        try:
            parse_string(input)
        except ParserError as e:
            assert 'The "build" field is no longer supported.' in e.message
        else:
            assert False, 'expected ParserError'

    def test_name_prefix(self):
        input = dedent('''\
            git module foo:
                url: fun stuff

            rule bar:
                export: more stuff
            ''')
        scope, imports = parse_string(input, name_prefix='x')
        # Lookup keys should be unaffected, but the names that modules and
        # rules give for themselves should have the prefix.
        assert scope.modules['foo'].name == 'xfoo'
        assert scope.rules['bar'].name == 'xbar'

    def test_forgotten_colon(self):
        # There are many different permutations of this error, and this only
        # tests the one mentioned in
        # https://github.com/keybase/client/issues/242.
        # TODO: A more general data validation library might help the parser do
        # a better job of checking these things. See
        # https://github.com/buildinspace/peru/issues/40.
        input = dedent('''\
            rule test:
                pick bla
            ''')
        with self.assertRaises(ParserError):
            parse_string(input)

    def test_duplicate_key_heuristic(self):
        yaml = dedent('''\
            a:
                a: 1
                b: 1
            b:
                a: 1
                b: 1
                a: 1
            a : whitespace before colon
            a: stuff
        ''')
        duplicates = parser._get_duplicate_keys_approximate(yaml)
        self.assertEqual([
            ('a', 5, 7),
            ('a', 1, 8),
            ('a', 8, 9),
        ], duplicates)



import hashlib
import importlib.util
import io
from os.path import abspath, join, dirname
import urllib

import peru
import shared

# https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
curl_plugin_path = abspath(
    join(
        dirname(peru.__file__), 'resources', 'plugins', 'curl',
        'curl_plugin.py'))
spec = importlib.util.spec_from_file_location("curl_plugin", curl_plugin_path)
curl_plugin = importlib.util.module_from_spec(spec)
spec.loader.exec_module(curl_plugin)


class MockRequest:
    def __init__(self, url, info, response):
        self.url = url
        self._info = info
        self._response_buffer = io.BytesIO(response)

    def info(self):
        return self._info

    def read(self, *args):
        return self._response_buffer.read(*args)


class CurlPluginTest(shared.PeruTest):
    def test_format_bytes(self):
        self.assertEqual('0B', curl_plugin.format_bytes(0))
        self.assertEqual('999B', curl_plugin.format_bytes(999))
        self.assertEqual('1.0KB', curl_plugin.format_bytes(1000))
        self.assertEqual('999.9KB', curl_plugin.format_bytes(999999))
        self.assertEqual('1.0MB', curl_plugin.format_bytes(10**6))
        self.assertEqual('1.0GB', curl_plugin.format_bytes(10**9))
        self.assertEqual('1000.0GB', curl_plugin.format_bytes(10**12))

    def test_get_request_filename(self):
        request = MockRequest('http://www.example.com/', {}, b'junk')
        self.assertEqual('index.html',
                         curl_plugin.get_request_filename(request))
        request.url = 'http://www.example.com/foo'
        self.assertEqual('foo', curl_plugin.get_request_filename(request))
        request._info = {'Content-Disposition': 'attachment; filename=bar'}
        self.assertEqual('bar', curl_plugin.get_request_filename(request))
        # Check quoted filenames.
        request._info = {'Content-Disposition': 'attachment; filename="bar"'}
        self.assertEqual('bar', curl_plugin.get_request_filename(request))
        # Check backslashed quotes in filenames.
        request._info = {
            'Content-Disposition': 'attachment; filename="bar\\""'
        }
        self.assertEqual('bar"', curl_plugin.get_request_filename(request))

    def test_download_file_with_length(self):
        content = b'xy' * 4096
        request = MockRequest('some url', {'Content-Length': len(content)},
                              content)
        stdout = io.StringIO()
        output_file = io.BytesIO()
        sha1 = curl_plugin.download_file(request, output_file, stdout)
        self.assertEqual(
            'downloaded 50% 4.0KB/8.1KB\ndownloaded 100% 8.1KB/8.1KB\n',
            stdout.getvalue())
        self.assertEqual(content, output_file.getvalue())
        self.assertEqual(hashlib.sha1(content).hexdigest(), sha1)

    def test_download_file_without_length(self):
        content = b'foo'
        request = MockRequest('some url', {}, content)
        stdout = io.StringIO()
        output_file = io.BytesIO()
        sha1 = curl_plugin.download_file(request, output_file, stdout)
        self.assertEqual('downloaded 3B\n', stdout.getvalue())
        self.assertEqual(content, output_file.getvalue())
        self.assertEqual(hashlib.sha1(content).hexdigest(), sha1)

    def test_unpack_windows_zip(self):
        '''This zip was packed on Windows, so it doesn't include any file
        permissions. This checks that our executable-flag-restoring code
        doesn't barf when the flag isn't there.'''
        test_dir = shared.create_dir()
        archive = shared.test_resources / 'from_windows.zip'
        curl_plugin.extract_zip(str(archive), test_dir)
        shared.assert_contents(test_dir, {'windows_test/test.txt': 'Notepad!'})
        txt_file = join(test_dir, 'windows_test/test.txt')
        shared.assert_not_executable(txt_file)

    def test_evil_archives(self):
        '''Even though most zip and tar utilities try to prevent absolute paths
        and paths starting with '..', it's entirely possible to construct an
        archive with either. These should always be an error.'''
        dest = shared.create_dir()
        for case in 'absolute_path', 'leading_dots':
            zip_archive = shared.test_resources / (case + '.zip')
            with self.assertRaises(curl_plugin.EvilArchiveError):
                curl_plugin.extract_zip(str(zip_archive), dest)
            tar_archive = shared.test_resources / (case + '.tar')
            with self.assertRaises(curl_plugin.EvilArchiveError):
                curl_plugin.extract_tar(str(tar_archive), dest)

    def test_evil_symlink_archives(self):
        """Even worse than archives containing bad paths, an archive could
        contain a *symlink* pointing to a bad path. Then a subsequent entry in
        the *same* archive could write through the symlink."""
        dest = shared.create_dir()
        for case in ["illegal_symlink_dots", "illegal_symlink_absolute"]:
            tar_archive = shared.test_resources / (case + ".tar")
            with self.assertRaises(curl_plugin.EvilArchiveError):
                curl_plugin.extract_tar(str(tar_archive), dest)
        # But leading dots should be allowed in symlinks, as long as they don't
        # escape the root of the archive.
        for case in ["legal_symlink_dots"]:
            tar_archive = shared.test_resources / (case + ".tar")
            curl_plugin.extract_tar(str(tar_archive), dest)

    def test_request_has_user_agent_header(self):
        actual = curl_plugin.build_request("http://example.test")
        self.assertTrue(actual.has_header("User-agent"))
        ua_header = actual.get_header("User-agent")
        peru_component, urllib_component = ua_header.split(' ')
        _, peru_version = peru_component.split('/')
        _, urllib_version = urllib_component.split('/')
        self.assertEqual(peru.main.get_version(), peru_version)
        self.assertEqual(urllib.request.__version__, urllib_version)



from peru.cache import Cache
from peru.merge import merge_imports_tree

from shared import create_dir, assert_contents, PeruTest, make_synchronous


class MergeTest(PeruTest):
    @make_synchronous
    async def setUp(self):
        self.cache_dir = create_dir()
        self.cache = await Cache(self.cache_dir)

        # These tests use this simple one-file tree as module contents.
        content = {'a': 'a'}
        content_dir = create_dir(content)
        self.content_tree = await self.cache.import_tree(content_dir)

    @make_synchronous
    async def test_merge_from_map(self):
        imports = {'foo': ('path1', ), 'bar': ('path2', )}
        target_trees = {'foo': self.content_tree, 'bar': self.content_tree}

        merged_tree = await merge_imports_tree(self.cache, imports,
                                               target_trees)

        merged_dir = create_dir()
        await self.cache.export_tree(merged_tree, merged_dir)
        expected_content = {'path1/a': 'a', 'path2/a': 'a'}
        assert_contents(merged_dir, expected_content)

    @make_synchronous
    async def test_merge_from_multimap(self):
        # This represents a list of key-value pairs in YAML, for example:
        #     imports:
        #         foo:
        #           - path1
        #           - path2
        imports = {'foo': ('path1', 'path2')}
        target_trees = {'foo': self.content_tree}

        merged_tree = await merge_imports_tree(self.cache, imports,
                                               target_trees)

        merged_dir = create_dir()
        await self.cache.export_tree(merged_tree, merged_dir)
        expected_content = {'path1/a': 'a', 'path2/a': 'a'}
        assert_contents(merged_dir, expected_content)



#! /usr/bin/env python3

import hashlib
import os
import pathlib
import re
import stat
import sys
import tarfile
from urllib.error import HTTPError, URLError
from urllib.parse import urlsplit
from urllib.request import Request
import peru.main
import urllib.request
import zipfile


def add_user_agent_to_request(request):
    components = [
        "peru/%s" % peru.main.get_version(),
        urllib.request.URLopener.version
    ]
    request.add_header("User-agent", " ".join(components))
    return request


def build_request(url):
    request = Request(url)
    return add_user_agent_to_request(request)


def get_request_filename(request):
    '''Figure out the filename for an HTTP download.'''
    # Check to see if a filename is specified in the HTTP headers.
    if 'Content-Disposition' in request.info():
        disposition = request.info()['Content-Disposition']
        pieces = re.split(r'\s*;\s*', disposition)
        for piece in pieces:
            if piece.startswith('filename='):
                filename = piece[len('filename='):]
                # Strip exactly one " from each end.
                if filename.startswith('"'):
                    filename = filename[1:]
                if filename.endswith('"'):
                    filename = filename[:-1]
                # Interpret backslashed quotes.
                filename = filename.replace('\\"', '"')
                return filename
    # If no filename was specified, pick a reasonable default.
    return os.path.basename(urlsplit(request.url).path) or 'index.html'


def format_bytes(num_bytes):
    for threshold, unit in ((10**9, 'GB'), (10**6, 'MB'), (10**3, 'KB')):
        if num_bytes >= threshold:
            # Truncate floats instead of rounding.
            float_str = str(num_bytes / threshold)
            decimal_index = float_str.index('.')
            truncated_float = float_str[:decimal_index + 2]
            return truncated_float + unit
    return '{}B'.format(num_bytes)


def download_file(request, output_file, stdout=sys.stdout):
    digest = hashlib.sha1()
    file_size_str = request.info().get('Content-Length')
    file_size = int(file_size_str) if file_size_str is not None else None
    bytes_read = 0
    while True:
        buf = request.read(4096)
        if not buf:
            break
        digest.update(buf)
        if output_file:
            output_file.write(buf)
        bytes_read += len(buf)
        percentage = ''
        kb_downloaded = format_bytes(bytes_read)
        total_kb = ''
        if file_size:
            percentage = ' {}%'.format(round(100 * bytes_read / file_size))
            total_kb = '/' + format_bytes(file_size)
        print(
            'downloaded{} {}{}'.format(percentage, kb_downloaded, total_kb),
            file=stdout)
    return digest.hexdigest()


def plugin_sync(url, sha1):
    unpack = os.environ['PERU_MODULE_UNPACK']
    dest = os.environ['PERU_SYNC_DEST']
    if unpack:
        # Download to the tmp dir for later unpacking.
        download_dir = os.environ['PERU_PLUGIN_TMP']
    else:
        # Download directly to the destination dir.
        download_dir = dest

    with urllib.request.urlopen(build_request(url)) as request:
        filename = os.environ['PERU_MODULE_FILENAME']
        if not filename:
            filename = get_request_filename(request)
        full_filepath = os.path.join(download_dir, filename)
        with open(full_filepath, 'wb') as output_file:
            digest = download_file(request, output_file)

    if sha1 and digest != sha1:
        print(
            'Bad checksum!\n     url: {}\nexpected: {}\n  actual: {}'.format(
                url, sha1, digest),
            file=sys.stderr)
        sys.exit(1)

    try:
        if unpack == 'tar':
            extract_tar(full_filepath, dest)
        elif unpack == 'zip':
            extract_zip(full_filepath, dest)
        elif unpack:
            print('Unknown value for "unpack":', unpack, file=sys.stderr)
            sys.exit(1)
    except EvilArchiveError as e:
        print(e.message, file=sys.stderr)
        sys.exit(1)


def extract_tar(archive_path, dest):
    with tarfile.open(archive_path) as t:
        for info in t.getmembers():
            validate_filename(info.path)
            if info.issym():
                validate_symlink(info.path, info.linkname)
        # Python 3.12 added the `filter` kwarg, which should make our
        # validation redundant. (It was also added to patch releases of earlier
        # Python versions.) Python 3.13 made it a warning to omit this
        # argument, because Python 3.14 will change the default to "data".
        # That's the behavior we want, and specifying it here lets us get it on
        # Python 3.12/3.13 and silences the warning.
        kwargs = {}
        if sys.version_info >= (3, 12):
            kwargs["filter"] = "data"
        t.extractall(dest, **kwargs)


def extract_zip(archive_path, dest):
    with zipfile.ZipFile(archive_path) as z:
        for name in z.namelist():
            validate_filename(name)
        z.extractall(dest)
        # Set file permissions. Tar does this by default, but with zip we need
        # to do it ourselves.
        for info in z.filelist:
            if not info.filename.endswith('/'):
                # This is how to get file permissions out of a zip archive,
                # according to http://stackoverflow.com/q/434641/823869 and
                # http://bugs.python.org/file34873/issue15795_cleaned.patch.
                mode = (info.external_attr >> 16) & 0o777
                # Don't copy the whole mode, just set the executable bit. Two
                # reasons for this. 1) This is all going to end up in a git
                # tree, which only records the executable bit anyway. 2) Zip's
                # support for Unix file modes is nonstandard, so the mode field
                # is often zero and could be garbage. Mistakenly setting a file
                # executable isn't a big deal, but e.g. removing read
                # permissions would cause an error.
                if mode & stat.S_IXUSR:
                    os.chmod(os.path.join(dest, info.filename), 0o755)


def validate_filename(name):
    path = pathlib.PurePosixPath(name)
    if path.is_absolute() or ".." in path.parts:
        raise EvilArchiveError("Illegal path in archive: " + name)


def validate_symlink(name, target):
    # We might do this twice but that's fine.
    validate_filename(name)

    allowed_parent_parts = len(pathlib.PurePosixPath(name).parts) - 1

    target_path = pathlib.PurePosixPath(target)
    if target_path.is_absolute():
        raise EvilArchiveError("Illegal symlink target in archive: " + target)
    leading_parent_parts = 0
    for part in target_path.parts:
        if part != "..":
            break
        leading_parent_parts += 1
    if leading_parent_parts > allowed_parent_parts:
        raise EvilArchiveError("Illegal symlink target in archive: " + target)
    if ".." in target_path.parts[leading_parent_parts:]:
        raise EvilArchiveError("Illegal symlink target in archive: " + target)


class EvilArchiveError(RuntimeError):
    def __init__(self, message):
        self.message = message


def plugin_reup(url, sha1):
    reup_output = os.environ['PERU_REUP_OUTPUT']
    with urllib.request.urlopen(build_request(url)) as request:
        digest = download_file(request, None)
    with open(reup_output, 'w') as output_file:
        print('sha1:', digest, file=output_file)


def main():
    url = os.environ['PERU_MODULE_URL']
    sha1 = os.environ['PERU_MODULE_SHA1']
    command = os.environ['PERU_PLUGIN_COMMAND']
    try:
        if command == 'sync':
            plugin_sync(url, sha1)
        elif command == 'reup':
            plugin_reup(url, sha1)
        else:
            raise RuntimeError('unknown command: ' + repr(command))
    except (HTTPError, URLError) as e:
        print("Error fetching", url)
        print(e)
        return 1


if __name__ == '__main__':
    sys.exit(main())



import os
import shutil
import textwrap

from peru.compat import makedirs
from peru.runtime import CommandLineError

import shared


class PathsTest(shared.PeruTest):
    def setUp(self):
        self.test_root = shared.create_dir()

        self.project_dir = os.path.join(self.test_root, 'project')
        self.state_dir = os.path.join(self.project_dir, '.peru')
        self.cache_dir = os.path.join(self.state_dir, 'cache')

        self.yaml = textwrap.dedent('''\
            imports:
                foo: ./
            cp module foo:
                # relative paths should always be interpreted from the dir
                # containing peru.yaml, even if that's not the sync dir.
                path: ../foo
            ''')
        shared.write_files(self.project_dir, {'peru.yaml': self.yaml})
        self.peru_file = os.path.join(self.project_dir, 'peru.yaml')

        self.foo_dir = os.path.join(self.test_root, 'foo')
        shared.write_files(self.foo_dir, {'bar': 'baz'})

        # We'll run tests from this inner subdirectory, so that we're more
        # likely to catch places where we're using the cwd incorrectly.
        self.cwd = os.path.join(self.project_dir, 'cwd', 'in', 'here')
        makedirs(self.cwd)

    def assert_success(self, sync_dir, state_dir, cache_dir, more_excludes=[]):
        shared.assert_contents(
            sync_dir, {'bar': 'baz'},
            excludes=['.peru', 'peru.yaml'] + more_excludes)
        assert os.path.isfile(os.path.join(state_dir, 'lastimports'))
        assert os.path.isdir(os.path.join(cache_dir, 'trees'))

    def test_unmodified_sync(self):
        shared.run_peru_command(['sync'], self.cwd)
        self.assert_success(self.project_dir, self.state_dir, self.cache_dir)

    def test_peru_file_and_sync_dir_must_be_set_together(self):
        for command in [['--sync-dir=junk', 'sync'], ['--file=junk', 'sync']]:
            with self.assertRaises(CommandLineError):
                shared.run_peru_command(command, cwd=self.cwd)

    def test_file_and_file_basename_incompatible(self):
        with self.assertRaises(CommandLineError):
            shared.run_peru_command([
                '--file=foo', '--sync-dir=bar', '--file-basename=baz', 'sync'
            ],
                                    cwd=self.cwd)

    def test_setting_all_flags(self):
        cwd = shared.create_dir()
        sync_dir = shared.create_dir()
        state_dir = shared.create_dir()
        cache_dir = shared.create_dir()
        shared.run_peru_command([
            '--file', self.peru_file, '--sync-dir', sync_dir, '--state-dir',
            state_dir, '--cache-dir', cache_dir, 'sync'
        ], cwd)
        self.assert_success(sync_dir, state_dir, cache_dir)

    def test_setting_all_env_vars(self):
        cache_dir = shared.create_dir()
        shared.run_peru_command(['sync'],
                                self.cwd,
                                env={
                                    'PERU_CACHE_DIR': cache_dir,
                                })
        self.assert_success(self.project_dir, self.state_dir, cache_dir)

    def test_flags_override_vars(self):
        flag_cache_dir = shared.create_dir()
        env_cache_dir = shared.create_dir()
        shared.run_peru_command(['--cache-dir', flag_cache_dir, 'sync'],
                                self.cwd,
                                env={'PERU_CACHE_DIR': env_cache_dir})
        self.assert_success(self.project_dir, self.state_dir, flag_cache_dir)

    def test_relative_paths(self):
        '''We ran into a bug where calling os.path.dirname(peru_file) was
        returning "", which got passed as the cwd of a plugin job and blew up.
        This test repros that case. We've switched to pathlib.Path.parent to
        fix the issue.'''
        shared.run_peru_command(
            ['--file', 'peru.yaml', '--sync-dir', '.', 'sync'],
            cwd=self.project_dir)
        self.assert_success(self.project_dir, self.state_dir, self.cache_dir)

    def test_default_file_name(self):
        shutil.move(self.peru_file, os.path.join(self.project_dir, 'xxx'))
        shared.run_peru_command(['--file-basename', 'xxx', 'sync'],
                                cwd=self.cwd)
        self.assert_success(
            self.project_dir,
            self.state_dir,
            self.cache_dir,
            more_excludes=['xxx'])



import os
import setuptools
import sys

# Written according to the docs at
# https://packaging.python.org/en/latest/distributing.html

project_root = os.path.dirname(__file__)
readme_file = os.path.join(project_root, 'README.md')
module_root = os.path.join(project_root, 'peru')
version_file = os.path.join(module_root, 'VERSION')


def get_version():
    with open(version_file) as f:
        return f.read().strip()


def get_all_resources_filepaths():
    resources_paths = ['VERSION']
    resources_dir = os.path.join(module_root, 'resources')
    for dirpath, dirnames, filenames in os.walk(resources_dir):
        relpaths = [
            os.path.relpath(os.path.join(dirpath, f), start=module_root)
            for f in filenames
        ]
        resources_paths.extend(relpaths)
    return resources_paths


def get_install_requires():
    dependencies = ['PyYAML']
    if sys.version_info < (3, 5):
        raise RuntimeError('The minimum supported Python version is 3.5.')
    return dependencies


def readme_text():
    with open(readme_file) as f:
        return f.read().strip()


setuptools.setup(
    name='peru',
    description='A tool for fetching code',
    version=get_version(),
    url='https://github.com/buildinspace/peru',
    author="Jack O'Connor <oconnor663@gmail.com>, "
    "Sean Olson <olson.sean.k@gmail.com>",
    license='MIT',
    packages=['peru', 'peru.docopt'],
    package_data={'peru': get_all_resources_filepaths()},
    entry_points={'console_scripts': [
        'peru=peru.main:main',
    ]},
    install_requires=get_install_requires(),
    long_description=readme_text(),
    long_description_content_type='text/markdown',
)



import asyncio
from asyncio.subprocess import PIPE
import sys

from peru.async_helpers import safe_communicate
from shared import PeruTest, make_synchronous


class AsyncTest(PeruTest):
    @make_synchronous
    async def test_safe_communicate(self):
        # Test safe_communicate with both empty and non-empty input.
        cat_command = [
            sys.executable, "-c",
            "import sys; sys.stdout.write(sys.stdin.read())"
        ]

        proc_empty = await asyncio.create_subprocess_exec(
            *cat_command, stdin=PIPE, stdout=PIPE)
        stdout, _ = await safe_communicate(proc_empty, b"")
        self.assertEqual(stdout, b"")

        proc_nonempty = await asyncio.create_subprocess_exec(
            *cat_command, stdin=PIPE, stdout=PIPE)
        stdout, _ = await safe_communicate(proc_nonempty, b"foo bar baz")
        self.assertEqual(stdout, b"foo bar baz")

        # And test a case with None input as well.
        true_command = [sys.executable, "-c", ""]
        proc_true = await asyncio.create_subprocess_exec(
            *true_command, stdin=PIPE, stdout=PIPE)
        stdout, _ = await safe_communicate(proc_true)
        self.assertEqual(stdout, b"")



from textwrap import dedent

from peru import plugin

import shared


def assert_parallel(n):
    # The plugin module keep a global counter of all the jobs that run in
    # parallel, so that we can write these tests.
    if plugin.DEBUG_PARALLEL_MAX != n:
        raise AssertionError('Expected {} parallel {}. Counted {}.'.format(
            n, 'job' if n == 1 else 'jobs', plugin.DEBUG_PARALLEL_MAX))


class ParallelismTest(shared.PeruTest):
    def setUp(self):
        # Make sure nothing is fishy with the jobs counter, and reset the max.
        plugin.debug_assert_clean_parallel_count()
        plugin.DEBUG_PARALLEL_MAX = 0

    def tearDown(self):
        # Make sure nothing is fishy with the jobs counter. No sense in
        # resetting the max here, because the rest of our tests don't know to
        # reset it anyway.
        plugin.debug_assert_clean_parallel_count()

    def test_two_jobs_in_parallel(self):
        # This just checks that two different modules can actually be fetched
        # in parallel.
        foo = shared.create_dir()
        bar = shared.create_dir()
        peru_yaml = dedent('''\
            imports:
                foo: ./
                bar: ./

            cp module foo:
                path: {}

            cp module bar:
                path: {}
            '''.format(foo, bar))
        test_dir = shared.create_dir({'peru.yaml': peru_yaml})
        shared.run_peru_command(['sync'], test_dir)
        assert_parallel(2)

    def test_jobs_flag(self):
        # This checks that the --jobs flag is respected, even when two modules
        # could have been fetched in parallel.
        foo = shared.create_dir()
        bar = shared.create_dir()
        peru_yaml = dedent('''\
            imports:
                foo: ./
                bar: ./

            cp module foo:
                path: {}

            cp module bar:
                path: {}
            '''.format(foo, bar))
        test_dir = shared.create_dir({'peru.yaml': peru_yaml})
        shared.run_peru_command(['sync', '-j1'], test_dir)
        assert_parallel(1)

    def test_identical_fields(self):
        # This checks that modules with identical fields are not fetched in
        # parallel. This is the same logic that protects us from fetching a
        # given module twice, like when it's imported with two different named
        # rules.
        foo = shared.create_dir()
        peru_yaml = dedent('''\
            imports:
                foo1: ./
                foo2: ./

            cp module foo1:
                path: {}

            cp module foo2:
                path: {}
            '''.format(foo, foo))
        test_dir = shared.create_dir({'peru.yaml': peru_yaml})
        shared.run_peru_command(['sync'], test_dir)
        assert_parallel(1)

    def test_identical_plugin_cache_fields(self):
        # Plugins that use caching also need to avoid running in parallel, if
        # their cache directories are the same. The noop_cache plugin (created
        # for this test) uses the path field (but not the nonce field) in its
        # plugin cache key. Check that these two modules are not fetched in
        # parallel, even though their module fields aren't exactly the same.
        foo = shared.create_dir()
        peru_yaml = dedent('''\
            imports:
                foo1: ./
                foo2: ./

            noop_cache module foo1:
                path: {}
                # nonce is ignored, but it makes foo1 different from foo2 as
                # far as the module cache is concerned
                nonce: '1'

            noop_cache module foo2:
                path: {}
                nonce: '2'
            '''.format(foo, foo))
        test_dir = shared.create_dir({'peru.yaml': peru_yaml})
        shared.run_peru_command(['sync'], test_dir)
        assert_parallel(1)



import importlib.machinery
from os.path import abspath, join, dirname

import peru
import shared

# https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
git_plugin_path = abspath(
    join(
        dirname(peru.__file__), 'resources', 'plugins', 'git',
        'git_plugin.py'))
spec = importlib.util.spec_from_file_location("git_plugin", git_plugin_path)
git_plugin = importlib.util.module_from_spec(spec)
spec.loader.exec_module(git_plugin)


# NOTE: The sync/reup functionality for the git plugin is tested in
# test_plugins.py along with the other plugin types.
class GitPluginTest(shared.PeruTest):
    def test_expand_relative_submodule_url(self):
        cases = [
            ("http://foo.com/a/b", "c", "c"),
            ("http://foo.com/a/b", "./c", "http://foo.com/a/b/./c"),
            ("http://foo.com/a/b", "../c", "http://foo.com/a/b/../c"),
            ("http://foo.com/a/b", "../../c", "http://foo.com/a/b/../../c"),
            ("http://foo.com/a/b", ".//../c", "http://foo.com/a/b/.//../c"),
        ]
        for (parent, submodule, expected) in cases:
            result = git_plugin.expand_relative_submodule_url(
                submodule, parent)
            assert expected == result, "{} != {}".format(expected, result)



import os

import peru.compat as compat
import shared


class CompatTest(shared.PeruTest):
    def test_makedirs(self):
        tmp_dir = shared.tmp_dir()
        foo_dir = os.path.join(tmp_dir, "foo")
        compat.makedirs(foo_dir)
        os.chmod(foo_dir, 0o700)
        # Creating the dir again should be a no-op even though the permissions
        # have changed.
        compat.makedirs(foo_dir)



import io
import re
import textwrap

from peru import display
import shared


class DisplayTest(shared.PeruTest):
    def test_quiet_display(self):
        output = io.StringIO()
        disp = display.QuietDisplay(output)
        with disp.get_handle('title') as handle:
            handle.write('some stuff!')
        disp.print('other stuff?')
        self.assertEqual('other stuff?\n', output.getvalue())

    def test_verbose_display(self):
        output = io.StringIO()
        disp = display.VerboseDisplay(output)
        with disp.get_handle('title') as handle:
            handle.write('in job 1\n')
            disp.print('print stuff')
            handle.write('in job 2\n')
        expected = textwrap.dedent('''\
            === started title ===
            print stuff
            === finished title ===
            in job 1
            in job 2
            ===
            ''')
        self.assertEqual(expected, output.getvalue())

    def test_fancy_display(self):
        output = FakeTerminal()
        disp = display.FancyDisplay(output)

        handle1 = disp.get_handle('title1')
        handle1.__enter__()
        handle1.write('something1')
        disp._draw()
        # We need to test trailing spaces, and the '# noqa: W291' tag stops the
        # linter from complaining about these.
        expected1 = textwrap.dedent('''\
            ╶ title1: something1 
            ''')  # noqa: W291
        self.assertEqual(expected1, output.getlines())

        handle2 = disp.get_handle('title2')
        handle2.__enter__()
        handle2.write('something2')
        disp._draw()
        expected2 = textwrap.dedent('''\
            ┌ title1: something1 
            └ title2: something2 
            ''')  # noqa: W291
        self.assertEqual(expected2, output.getlines())

        handle3 = disp.get_handle('title3')
        handle3.__enter__()
        handle3.write('something3')
        disp._draw()
        expected3 = textwrap.dedent('''\
            ┌ title1: something1 
            ├ title2: something2 
            └ title3: something3 
            ''')  # noqa: W291
        self.assertEqual(expected3, output.getlines())

        disp.print('stuff above')
        # Calling _draw() should not be necessary after print(). This ensures
        # that we won't lose output if the program exits before _draw_later()
        # gets another chance to fire.
        expected4 = textwrap.dedent('''\
            stuff above
            ┌ title1: something1 
            ├ title2: something2 
            └ title3: something3 
            ''')  # noqa: W291
        self.assertEqual(expected4, output.getlines())

        handle2.__exit__(None, None, None)
        disp._draw()
        expected5 = textwrap.dedent('''\
            stuff above
            ┌ title1: something1 
            └ title3: something3 
            ''')  # noqa: W291
        self.assertEqual(expected5, output.getlines())

        handle1.__exit__(None, None, None)
        disp._draw()
        expected6 = textwrap.dedent('''\
            stuff above
            ╶ title3: something3 
            ''')  # noqa: W291
        self.assertEqual(expected6, output.getlines())

        handle3.__exit__(None, None, None)
        # _draw() should not be necessary after the last job exits.
        expected7 = textwrap.dedent('''\
            stuff above
            ''')
        self.assertEqual(expected7, output.getlines())
        self.assertEqual(None, disp._draw_later_handle)


class FakeTerminal:
    '''Emulates a terminal by keeping track of a list of lines. Knows how to
    interpret the ANSI escape sequences that are used by FancyDisplay.'''

    def __init__(self):
        self.lines = [io.StringIO()]
        self.cursor_line = 0
        # Flush doesn't actually do anything in fake terminal, but we want to
        # make sure it gets called before any lines are read.
        self.flushed = False

    def write(self, string):
        tokens = [
            display.ANSI_DISABLE_LINE_WRAP, display.ANSI_ENABLE_LINE_WRAP,
            display.ANSI_CLEAR_LINE, display.ANSI_CURSOR_UP_ONE_LINE, '\n'
        ]
        # The parens make this a capturing expression, so the tokens will be
        # included in re.split()'s return list.
        token_expr = '(' + '|'.join(re.escape(token) for token in tokens) + ')'
        pieces = re.split(token_expr, string)

        for piece in pieces:
            if piece in (display.ANSI_DISABLE_LINE_WRAP,
                         display.ANSI_ENABLE_LINE_WRAP):
                # Ignore the line wrap codes. TODO: Test for these?
                continue
            elif piece == display.ANSI_CLEAR_LINE:
                buffer = self.lines[self.cursor_line]
                buffer.seek(0)
                buffer.truncate()
            elif piece == display.ANSI_CURSOR_UP_ONE_LINE:
                col = self.lines[self.cursor_line].tell()
                self.cursor_line -= 1
                assert self.cursor_line >= 0
                new_buffer = self.lines[self.cursor_line]
                new_buffer.seek(col)
            elif piece == '\n':
                self.cursor_line += 1
                if self.cursor_line == len(self.lines):
                    self.lines.append(io.StringIO())
                self.lines[self.cursor_line].seek(0)
            else:
                self.lines[self.cursor_line].write(piece)

    def flush(self):
        self.flushed = True

    def getlines(self):
        # Make sure flush() was called after the last write.
        assert self.flushed
        self.flushed = False
        # Make sure none of the lines at or beyond the cursor have any text.
        for i in range(self.cursor_line, len(self.lines)):
            assert self.lines[i].getvalue() == ''
        # Concatenate all the lines before the cursor, and append trailing
        # newlines.
        lines = io.StringIO()
        for i in range(self.cursor_line):
            lines.write(self.lines[i].getvalue())
            lines.write('\n')
        return lines.getvalue()
